###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

###############################################################################
# Imports
###############################################################################

import logging
import re

from pants.stream import Stream

###############################################################################
# Logging
###############################################################################

log = logging.getLogger('irc')

###############################################################################
# Constants
###############################################################################

__all__ = ('BaseIRC','IRCClient')

COMMAND = re.compile(r"((?::.+? )?)(.+?) (.*)")
NETMASK = re.compile(
    r":(?:(?:([^!\s]+)!)?([^@\s]+)@)?([A-Za-z0-9-/:?`[_^{|}\\\]\.]+)")
ARGS    = re.compile(r"(?:^|(?<= ))(:.*|[^ ]+)")
CTCP    = re.compile(r"([\x00\n\r\x10])")
unCTCP  = re.compile(r"\x10([0nr\x10])")

CODECS  = ('utf-8','iso-8859-1','cp1252')
#])" # this is here to correct syntax highlighting in textmate... remove it!!!

###############################################################################
# BaseIRC Class
###############################################################################

class BaseIRC(Stream):
    """
    The IRC protocol, implemented over a Pants :class:`~pants.stream.Stream`.

    The goal with this is to create a lightweight IRC class that can serve as
    either a server or a client. As such, it doesn't implement a lot of logic
    in favor of providing a robust base.

    The BaseIRC class can receive and send IRC commands, and automatically
    respond to certain commands such as PING.

    This class extends :class:`~pants.stream.Stream`, and as such has the same
    :func:`~pants.stream.Stream.connect` and :func:`~pants.stream.Stream.listen`
    functions.
    """
    def __init__(self, socket=None, encoding='utf-8'):
        Stream.__init__(self, socket=socket)

        # Set our prefix to an empty string. This is prepended to all sent
        # commands, and useful for servers.
        self.prefix = ''
        self.encoding = encoding

        # Read lines at once.
        self.read_delimiter = '\n'

    ##### Public Event Handlers ###############################################

    def irc_close(self):
        """
        Placeholder.

        This method is called whenever the IRC instance becomes disconnected
        from the remote client or server.
        """
        pass

    def irc_command(self, command, args, nick, user, host):
        """
        Placeholder.

        This method is called whenever a command is received from the other
        side and successfully parsed as an IRC command.

        =========  ============
        Argument   Description
        =========  ============
        command    The received command.
        args       A list of the arguments following the command.
        nick       The nick of the user that sent the command, if applicable, or an empty string.
        user       The username of the user that sent the command, if applicable, or an empty string.
        host       The host of the user that sent the command, the host of the server that sent the command, or an empty string if no host was supplied.
        =========  ============
        """
        pass

    def irc_connect(self):
        """
        Placeholder.

        This method is called when the IRC instance has successfully connected
        to the remote client or server.
        """
        pass

    ##### I/O Methods #########################################################

    def message(self, destination, message, _ctcpQuote=True, _prefix=None):
        """
        Send a message to the given nick or channel.

        ============  ========  ============
        Argument      Default   Description
        ============  ========  ============
        destination             The nick or channel to send the message to.
        message                 The text of the message to be sent.
        _ctcpQuote    True      *Optional.* If True, the message text will be quoted for CTCP before being sent.
        _prefix       None      *Optional.* A string that, if provided, will be prepended to the command string before it's sent to the server.
        ============  ========  ============
        """
        if _ctcpQuote:
            message = ctcpQuote(message)

        self.send_command("PRIVMSG", destination, message, _prefix=_prefix)

    def notice(self, destination, message, _ctcpQuote=True, _prefix=None):
        """
        Send a NOTICE to the specified destination.

        ===========  ========  ============
        Argument     Default   Description
        ===========  ========  ============
        destination            The nick or channel to send the NOTICE to.
        message                The text of the NOTICE to be sent.
        _ctcpQuote   True      *Optional.* If True, the message text will be quoted for CTCP before being sent.
        _prefix      None      *Optional.* A string that, if provided, will be prepended to the command string before it's sent to the server.
        ===========  ========  ============
        """
        if _ctcpQuote:
            message = ctcpQuote(message)

        self.send_command("NOTICE", destination, message, _prefix=_prefix)

    def quit(self, reason=None, _prefix=None):
        """
        Send a QUIT message, with an optional reason.

        =========  ========  ============
        Argument   Default   Description
        =========  ========  ============
        reason     None      *Optional.* The reason for quitting that will be displayed to other users.
        _prefix    None      *Optional.* A string that, if provided, will be prepended to the command string before it's sent to the server.
        =========  ========  ============
        """
        if not reason:
            reason = "pants.contrib.irc -- http://www.pantsweb.org/"
        self.send_command("QUIT", reason, _prefix=_prefix)

    def send_command(self, command, *args, **kwargs):
        """
        Send a command to the remote endpoint.

        =========  ========  ============
        Argument   Default   Description
        =========  ========  ============
        command              The command to send.
        \*args                *Optional.* A list of arguments to send with the command.
        _prefix    None      *Optional.* A string that, if provided, will be prepended to the command string before it's sent to the server.
        =========  ========  ============
        """
        if args:
            args = list(args)
            for i in xrange(len(args)):
                arg = args[i]
                if not isinstance(arg, basestring):
                    args[i] = str(arg)

            if not args[-1].startswith(':'):
                args[-1] = ':%s' % args[-1]

            out = '%s %s\r\n' % (command, ' '.join(args))
        else:
            out = '%s\r\n' % command

        if '_prefix' in kwargs and kwargs['_prefix']:
            out = '%s %s' % (kwargs['_prefix'], out)
        elif self.prefix:
            out = '%s %s' % (self.prefix, out)

        # Send it.
        log.debug('\x1B[0;32m>> %s\x1B[0m' % out.rstrip())

        self.write(out.encode(self.encoding))

    ##### Internal Event Handlers #############################################

    def on_command(self, command, args, nick, user, host):
        """
        Placeholder.

        Performs any logic that has to be performed upon receiving a command,
        then calls irc_command.

        Arguments are identical to irc_command.
        """
        if hasattr(self, 'irc_command_%s' % command):
            getattr(self, 'irc_command_%s' % command)(
                command, args, nick, user, host)
        else:
            self.irc_command(command, args, nick, user, host)

    def on_connect(self):
        """
        Placeholder.

        Performs any logic that has to be performed at connect, then calls
        self.irc_connect.
        """
        self.irc_connect()

    def on_close(self):
        """
        Placeholder.

        Performs any logic that has to be performed at disconnect, then calls
        self.irc_close.
        """
        self.irc_close()

    def on_read(self, data):
        """
        Read the available data, parse the command, and call an event for it.
        """
        data = data.strip('\r\n')
        if not data:
            return

        log.debug('\x1B[0;31m<< %s\x1B[0m' % repr(data)[1:-1])

        # Decode it straight away.
        data = decode(data)

        try:
            prefix, command, raw = COMMAND.match(data).groups()
        except Exception:
            log.warning('Invalid IRC command %r.' % data)
            return

        if prefix:
            nick, user, host = NETMASK.match(prefix).groups()
            if not nick and not '.' in host:
                nick = host
                host = ''
        else:
            nick, user, host = '', '', ''

        args = ARGS.findall(raw)
        if args:
            if args[-1].startswith(':'):
                args[-1] = args[-1][1:]

        # If it's PING, handle it.
        if command == 'PING':
            self.send_command('PONG', *args)
            return

        # Handle the command.
        self.on_command(command, args, nick, user, host)

###############################################################################
# IRCClient Class & Channel Class
###############################################################################

class Channel(object):
    """
    An IRC channel's representation, for keeping track of users and the topic
    and stuff.
    """
    __slots__ = ('name', 'users','topic','topic_setter','topic_time')

    def __init__(self, name):
        self.name = name
        self.users = []
        self.topic = None
        self.topic_setter = None
        self.topic_time = 0

class IRCClient(BaseIRC):
    """
    An IRC client, written in Pants, based on :class:`~pants.contrib.irc.BaseIRC`.

    This implements rather more logic, and keeps track of what server it's
    connected to, its nick, and what channels it's in.
    """
    def __init__(self, socket=None, encoding='utf-8'):
        BaseIRC.__init__(self, socket, encoding)

        # Internal State Stuff
        self._channels  = {}
        self._joining   = []

        self._nick      = None
        self._port      = 6667
        self._server    = None
        self._user      = None
        self._realname  = None

        # External Stuff
        self.password   = None

    ##### Properties ##########################################################

    @property
    def nick(self):
        """
        This instance's current nickname on the server it's connected to, or
        the nickname it will attempt to acquire when connecting.
        """
        return self._nick

    @nick.setter
    def nick(self, val):
        if not self.connected:
            self._nick = val
        else:
            self.send_command("NICK", val)

    @property
    def port(self):
        """
        The port this instance is connected to on the remote server, or the
        port it will attempt to connect to.
        """
        return self._port

    @port.setter
    def port(self, val):
        if self.connected or self.connecting:
            raise IOError('Cannot change while connected to server.')
        self._port = val

    @property
    def realname(self):
        """
        The real name this instance will report to the server when connecting.
        """
        return self._realname

    @realname.setter
    def realname(self, val):
        if self.connected or self.connecting:
            raise IOError('Cannot change while connected to server.')
        self._realname = val

    @property
    def server(self):
        """
        The server this instance is connected to, or will attempt to connect to.
        """
        return self._server

    @server.setter
    def server(self, val):
        if self.connected or self.connecting:
            raise IOError('Cannot change while connected to server.')
        self._server = val

    @property
    def user(self):
        """
        The user name this instance will report to the server when connecting.
        """
        return self._user

    @user.setter
    def user(self, val):
        if self.connected or self.connecting:
            raise IOError('Cannot change while connected to server.')
        self._user = val

    ##### General Methods #####################################################

    def channel(self, name):
        """
        Retrieve a Channel object for the channel ``name``, or None if we're
        not in that channel.
        """
        return self._channels.get(name, None)

    def connect(self, server=None, port=None):
        """
        Connect to the server.

        =========  ============
        Argument   Description
        =========  ============
        server     The host to connect to.
        port       The port to connect to on the remote server.
        =========  ============
        """
        if not self.connected and not self.connecting:
            if server:
                self._server = server
            if port:
                self._port = port


        Stream.connect(self, (self._server, self._port))

    ##### I/O Methods #########################################################

    def join(self, channel):
        """
        Join the specified channel.

        =========  ============
        Argument   Description
        =========  ============
        channel    The name of the channel to join.
        =========  ============
        """
        if channel in self._channels or channel in self._joining:
            return

        self._joining.append(channel)
        self.send_command("JOIN", channel)

    def part(self, channel, reason=None, force=False):
        """
        Leave the specified channel.

        =========  ========  ============
        Argument   Default   Description
        =========  ========  ============
        channel              The channel to leave.
        reason     None      *Optional.* The reason why.
        force      False     *Optional.* Don't ensure the client is actually *in* the named channel before sending ``PART``.
        =========  ========  ============
        """
        if not force and not channel in self._channels:
            return

        args = [channel]
        if reason:
            args.append(reason)

        self.send_command("PART", *args)

    ##### Public Event Handlers ###############################################

    def irc_ctcp(self, nick, message, user, host):
        """
        Placeholder.

        This method is called when the bot receives a CTCP message, which
        could, in theory, be anywhere in a PRIVMSG... annoyingly enough.

        =========  ============
        Argument   Description
        =========  ============
        nick       The nick of the user that sent the CTCP message, or an empty string if no nick is available.
        message    The full CTCP message.
        user       The username of the user that sent the CTCP message, or an empty string if no username is available.
        host       The host of the user that sent the CTCP message, or an empty string if no host is available.
        =========  ============
        """
        pass

    def irc_join(self, channel, nick, user, host):
        """
        Placeholder.

        This method is called when a user enters a channel. That also means
        that this function is called whenever this IRC client successfully
        joins a channel.

        =========  ============
        Argument   Description
        =========  ============
        channel    The channel a user has joined.
        nick       The nick of the user that joined the channel.
        user       The username of the user that joined the channel.
        host       The host of the user that joined the channel.
        =========  ============
        """
        pass

    def irc_message_channel(self, channel, message, nick, user, host):
        """
        Placeholder.

        This method is called when the client receives a message from a channel.

        =========  ============
        Argument   Description
        =========  ============
        channel    The channel the message was received in.
        message    The text of the message.
        nick       The nick of the user that sent the message.
        user       The username of the user that sent the message.
        host       The host of the user that sent the message.
        =========  ============
        """
        pass

    def irc_message_private(self, nick, message, user, host):
        """
        Placeholder.

        This method is called when the client receives a message from a user.

        =========  ============
        Argument   Description
        =========  ============
        nick       The nick of the user that sent the message.
        message    The text of the message.
        user       The username of the user that sent the message.
        host       The host of the user that sent the message.
        =========  ============
        """
        pass

    def irc_nick_changed(self, nick):
        """
        Placeholder.

        This method is called when the client's nick on the network is changed
        for any reason.

        =========  ============
        Argument   Description
        =========  ============
        nick       The client's new nick.
        =========  ============
        """
        pass

    def irc_part(self, channel, reason, nick, user, host):
        """
        Placeholder.

        This method is called when a leaves enters a channel. That also means
        that this function is called whenever this IRC client leaves a
        channel.

        =========  ============
        Argument   Description
        =========  ============
        channel    The channel that the user has left.
        reason     The provided reason message, or an empty string if there is no message.
        nick       The nick of the user that left the channel.
        user       The username of the user that left the channel.
        host       The host of the user that left the channel.
        =========  ============
        """
        pass

    def irc_topic_changed(self, channel, topic):
        """
        Placeholder.

        This method is called when the topic of a channel changes.

        =========  ============
        Argument   Description
        =========  ============
        channel    The channel which had its topic changed.
        topic      The channel's new topic.
        =========  ============
        """
        pass

    ##### IRC Command Handlers ################################################

    def irc_command_004(self, command, args, nick, user, host):
        """ 004 - Registered
            Syntax:
                004 server ver usermode chanmode

            The 004 command is sent once we've registered successfully with the
            server and can proceed to do normal IRC things.
        """
        # Check our nick.
        n = args[0]
        if n != self._nick:
            self._nick = n
            self.irc_nick_changed(n)

        self.irc_connect()

    def irc_command_332(self, command, args, nick, user, host):
        """ 332 - Channel Topic
            Syntax:
                332 channel :topic
        """
        chan, topic = args[-2:]
        if chan in self._channels:
            self._channels[chan].topic = topic

        self.irc_topic_changed(chan, topic)

    def irc_command_333(self, command, args, nick, user, host):
        """ 333 - Channel Topic (Extended)
            Syntax:
                333 channel nickname time
        """
        chan, nickname, time = args[-3:]
        if chan in self._channels:
            self._channels[chan].topic_setter = nickname
            try:
                self._channels[chan].topic_time = int(time)
            except ValueError:
                pass

    def irc_command_353(self, command, args, nick, user, host):
        """ 353 - Users in Channel
            Syntax:
                353 = channel: names
        """
        chan, names = args[-2:]
        if chan in self._channels:
            for name in names.split(' '):
                while name[0] in '@+':
                    name = name[1:]
                if not name in self._channels[chan].users:
                    self._channels[chan].users.append(name)

    def irc_command_JOIN(self, command, args, nick, user, host):
        """
        Received whenever a user, including ourself, joins a channel.
        """
        chan = args[0]

        if nick == self._nick:
            if chan in self._joining:
                self._joining.remove(chan)
            if not chan in self._channels:
                self._channels[chan] = Channel(chan)

        if chan in self._channels:
            name = nick
            while name[0] in '@+':
                name = name[1:]
            if not name in self._channels[chan].users:
                self._channels[chan].users.append(name)

        self.irc_join(chan, nick, user, host)

    def irc_command_PART(self, command, args, nick, user, host):
        """
        Received whenever a user, including ourself, leaves a channel.
        """
        chan = args[0]

        if nick == self._nick:
            if chan in self._joining:
                self._joining.remove(chan)
            if chan in self._channels:
                del self._channels[chan]

        if chan in self._channels:
            name = nick
            while name[0] in '@+':
                name = name[1:]
            if name in self._channels[chan].users:
                self._channels[chan].users.remove(name)

        if len(args) < 2:
            args.append('')

        self.irc_part(chan, args[1], nick, user, host)

    def irc_command_PRIVMSG(self, command, args, nick, user, host):
        """
        The PRIVMSG command is the heart of IRC communications. This method
        will call either irc_message_channel, or irc_message_private depending
        on the recipient of the privmsg.
        """
        msg = args[1]
        while msg:
            ind = msg.find('\x01')
            if ind == -1:
                ind = len(msg)

            if ind > 0:
                message = msg[:ind]
                msg = msg[ind:]

                if args[0] == self._nick:
                    self.irc_message_private(nick, message, user, host)
                else:
                    self.irc_message_channel(
                        args[0], message, nick, user, host)

            if msg:
                msg = msg[1:]
                ind = msg.find('\x01')
                if ind == -1:
                    continue

                message = msg[:ind]
                msg = msg[ind+1:]

                self.irc_ctcp(nick, message, user, host)

    ##### Internal Event Handlers #############################################

    def on_connect(self):
        """
        We're connected, so send the login stuff.
        """
        if self.password:
            self.send_command("PASS", self.password)

        self.send_command("NICK", self._nick or 'PantsIRC')

        # Our user and realname.
        self.send_command("USER",
            self._user or 'PantsIRC',
            0, '*',
            self._realname or 'pants.contrib.irc'
        )

        # And now, we wait. Don't raise irc_connect until we get a message
        # letting us know our connection was accepted.

    def on_close(self):
        """
        We've been disconnected.
        """
        self._channels  = {}
        self._joining   = []

        self.irc_close()

###############################################################################
# Helper Functions
###############################################################################

def ctcpQuote(message):
    """
    Low-level quote a message, adhering to the CTCP guidelines.
    """
    return CTCP.sub(_ctcpQuoter, message)

def _ctcpQuoter(match):
    m = match.group(1)
    if m == '\x00':
        return '\x100'
    elif m == '\n':
        return '\x10n'
    elif m == '\r':
        return '\x10r'
    elif m == '\x10':
        return '\x10\x10'
    else:
        return m

def ctcpUnquote(message):
    """
    Low-level unquote a message, adhering to the CTCP guidelines.
    """
    return unCTCP.sub(_ctcpUnquoter, message)

def _ctcpUnquoter(match):
    m = match.group(1)
    if m == '0':
        return '\x00'
    elif m == 'n':
        return '\n'
    elif m == 'r':
        return '\r'
    elif m == '\x10':
        return '\x10'
    else:
        return m

def decode(data):
    for codec in CODECS:
        try:
            return data.decode(codec)
        except UnicodeDecodeError:
            continue
    return data.decode('utf-8', 'ignore')



###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################
"""
Various implementations of the platform-dependant sendfile() system
call.
"""

###############################################################################
# Imports
###############################################################################

import os
import socket
import sys

import ctypes
import ctypes.util


###############################################################################
# Constants
###############################################################################

SENDFILE_PLATFORMS = ("linux2", "darwin", "freebsd", "dragonfly")
SENDFILE_AMOUNT = 2 ** 16


###############################################################################
# Implementations
###############################################################################

def sendfile_fallback(sfile, channel, offset, nbytes):
    """
    Fallback implementation of ``sendfile()``.

    This is not a true implementation of the ``sendfile()`` system call,
    but rather a fallback option written in Python. It has the same
    ultimate effect, but is far slower than a native implementation.
    This function is only used as a last resort.

    =========  ============
    Argument   Description
    =========  ============
    sfile      The file to send.
    channel    The channel to write to.
    offset     The number of bytes to offset writing by.
    nbytes     The number of bytes of the file to write. If 0, all bytes will be written.
    =========  ============
    """
    if nbytes == 0:
        to_read = SENDFILE_AMOUNT
    else:
        to_read = min(nbytes, SENDFILE_AMOUNT)

    sfile.seek(offset)
    data = sfile.read(to_read)

    if len(data) == 0:
        return 0

    return channel._socket_send(data)

def sendfile_linux(sfile, channel, offset, nbytes):
    """
    Linux 2.x implementation of ``sendfile()``.

    =========  ============
    Argument   Description
    =========  ============
    sfile      The file to send.
    channel    The channel to write to.
    offset     The number of bytes to offset writing by.
    nbytes     The number of bytes of the file to write. If 0, all bytes will be written.
    =========  ============
    """
    # TODO Linux doesn't support an argument of 0 for nbytes. Implement
    #      a better solution.
    if nbytes == 0:
        nbytes = SENDFILE_AMOUNT

    _offset = ctypes.c_uint64(offset)

    result = _sendfile(channel.fileno, sfile.fileno(), _offset, nbytes)

    if result == -1:
        e = ctypes.get_errno()
        raise socket.error(e, os.strerror(e))

    return result

def sendfile_darwin(sfile, channel, offset, nbytes):
    """
    Darwin implementation of ``sendfile()``.

    =========  ============
    Argument   Description
    =========  ============
    sfile      The file to send.
    channel    The channel to write to.
    offset     The number of bytes to offset writing by.
    nbytes     The number of bytes of the file to write. If 0, all bytes will be written.
    =========  ============
    """
    _nbytes = ctypes.c_uint64(nbytes)

    result = _sendfile(sfile.fileno(), channel.fileno, offset, _nbytes,
                       None, 0)

    if result == -1:
        e = ctypes.get_errno()
        raise socket.error(e, os.strerror(e))

    return _nbytes.value

def sendfile_bsd(sfile, channel, offset, nbytes):
    """
    FreeBSD/Dragonfly implementation of ``sendfile()``.

    =========  ============
    Argument   Description
    =========  ============
    sfile      The file to send.
    channel    The channel to write to.
    offset     The number of bytes to offset writing by.
    nbytes     The number of bytes of the file to write. If 0, all bytes will be written.
    =========  ============
    """
    _nbytes = ctypes.c_uint64()

    result = _sendfile(sfile.fileno(), channel.fileno, offset, nbytes,
                       None, _nbytes, 0)

    if result == -1:
        e = ctypes.get_errno()
        raise socket.error(e, os.strerror(e))

    return _nbytes.value


###############################################################################
# Sendfile
###############################################################################

_sendfile = None
if sys.version_info >= (2, 6) and sys.platform in SENDFILE_PLATFORMS:
    _libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)
    if hasattr(_libc, "sendfile"):
        _sendfile = _libc.sendfile

sendfile = None
if _sendfile is None:
    sendfile = sendfile_fallback

elif sys.platform == "linux2":
    _sendfile.argtypes = (
            ctypes.c_int,  # socket
            ctypes.c_int,  # file
            ctypes.POINTER(ctypes.c_uint64), #  offset
            ctypes.c_size_t  # len
            )

    sendfile = sendfile_linux

elif sys.platform == "darwin":
    _sendfile.argtypes = (
            ctypes.c_int,  # file
            ctypes.c_int,  # socket
            ctypes.c_uint64, # offset
            ctypes.POINTER(ctypes.c_uint64),  # len
            ctypes.c_voidp,  # header/trailer
            ctypes.c_int  # flags
            )

    sendfile = sendfile_darwin

elif sys.platform in ("freebsd", "dragonfly"):
    _sendfile.argtypes = (
            ctypes.c_int,  # file
            ctypes.c_int,  # socket
            ctypes.c_uint64,  # offset
            ctypes.c_uint64,  # len
            ctypes.c_voidp,  # header/trailer
            ctypes.POINTER(ctypes.c_uint64),  # bytes sent
            ctypes.c_int  # flags
            )

    sendfile = sendfile_bsd



#!/usr/bin/env python

from distutils.core import setup

setup(
    name="pants",
    version="0.10.1",
    description="A lightweight framework for writing asynchronous network applications in Python.",
    author="Christopher Davis",
    author_email="chris@wtfrak.com",
    url="http://pantsweb.org/",
    download_url="https://github.com/ecdavis/pants/tarball/pants-0.10.1",
    packages=["pants", "pants.contrib", "pants.contrib.http", "pants.util"],
    classifiers=[
        "Programming Language :: Python :: 2.6",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Server",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    scripts=[
        "bin/pantsd.py",
        "bin/pants_dig.py",
        ],
    )



###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

###############################################################################
# Imports
###############################################################################

import base64
import hashlib
import logging
import struct

###############################################################################
# Constants
###############################################################################

CLOSE_REASONS = {
    1000: 'Normal Closure',
    1001: 'Server Going Away',
    1002: 'Protocol Error',
    1003: 'Unacceptable Data Type',
    1004: 'Frame Too Large',
    1005: 'No Status Code',
    1006: 'Abnormal Close',
    1007: 'Invalid UTF-8 Data'
}

# Handshake Key
WEBSOCKET_KEY = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"

# Frame Opcodes
FRAME_CONTINUATION = 0
FRAME_TEXT = 1
FRAME_BINARY = 2
FRAME_CLOSE = 8
FRAME_PING = 9
FRAME_PONG = 10

# Special read_delimiter Value
EntireMessage = object()

###############################################################################
# Logging
###############################################################################

log = logging.getLogger(__name__)

###############################################################################
# WebSocketConnection Class
###############################################################################

class WebSocketConnection(object):
    """
    An implementation of `WebSockets <http://en.wikipedia.org/wiki/WebSockets>`_
    using the HTTP server provided as :class:`pants.contrib.http.HTTPServer`
    for the initial handshake, and adhering to the Pants design standards.

    Using this class, you can code using the same ``read_delimiter`` and
    named functions you would use when coding with
    :class:`~pants.stream.Stream`. WebSocket negotiation and framing occurs
    transparently.

    =========  ============
    Argument   Description
    =========  ============
    request    The :class:`~pants.contrib.http.HTTPRequest` to begin negotiating WebSockets upon.
    =========  ============
    """
    protocols = None

    def __init__(self, request):
        # Store the request and play nicely with web.
        self._connection = request.connection
        request.auto_finish = False

        # Base State
        self.fileno = self._connection.fileno

        self.read_delimiter = EntireMessage
        self._recv_buffer = ""
        self._read_buffer = u""

        self.connected = False

        # Set remote/local addr early for on_handshake.
        self._update_addr()

        # First up, make sure we're dealing with an actual WebSocket request.
        # If we aren't, return a simple 403 forbidden page.
        fail = False
        headers = {}

        if not request.headers.get('Connection','').lower() == 'upgrade' and \
                not request.headers.get('Upgrade','').lower() == 'websocket':
            fail = True

        # It's a WebSocket. Rejoice. Make sure the handshake information is
        # all acceptible.
        elif not self._safely_call(self.on_handshake, request, headers):
            fail = True

        # Determine which version of WebSockets we're dealing with.
        if 'Sec-WebSocket-Version' in request.headers:
            # New WebSockets. Handshake.
            if not 'Sec-WebSocket-Key' in request.headers:
                fail = True
            else:
                accept = base64.b64encode(hashlib.sha1(
                    request.headers['Sec-WebSocket-Key'] + WEBSOCKET_KEY
                    ).digest())

                headers['Upgrade'] = 'websocket'
                headers['Connection'] = 'Upgrade'
                headers['Sec-WebSocket-Accept'] = accept

                self.version = int(request.headers['Sec-WebSocket-Version'])

        else:
            # Old WebSockets. Wut?
            self.version = 0
            self._headers = headers
            self._request = request

            self._connection.on_read = self._finish_handshake
            self._connection.on_close = self._con_close
            self._connection.on_write = self._con_write
            self._connection.read_delimiter = 8
            return

        if fail:
            request.send_status(403)
            headers.update({
                'Content-Type': 'text/plain',
                'Content-Length': '13'
                })
            request.send_headers(headers)
            request.send("403 Forbidden")
            request.finish()
            return

        # Still here? No fail! Send the handshake response, hook up event
        # handlers, and call on_connect.
        request.send_status(101)
        request.send_headers(headers)

        self._connection.on_read = self._con_read
        self._connection.on_close = self._con_close
        self._connection.on_write = self._con_write
        self._connection.read_delimiter = None

        self.connected = True
        self._safely_call(self.on_connect)

    def _finish_handshake(self, key3):
        self._connection.read_delimiter = None
        request = self._request
        headers = self._headers
        del self._headers
        del self._request

        if request.protocol == 'https':
            scheme = 'wss'
        else:
            scheme = 'ws'

        request.send_status(101)
        headers.update({
            'Upgrade': 'WebSocket',
            'Connection': 'Upgrade',
            'Sec-WebSocket-Origin': request.headers['Origin'],
            'Sec-WebSocket-Location': '%s://%s%s' % (
                scheme, request.host, request.uri)
            })
        request.send_headers(headers)

        try:
            request.send(challenge_response(
                request.headers, key3))
        except ValueError:
            log.warning("Malformed WebSocket challenge to %s #%d." %
                (self.__class__.__name__, self.fileno))
            self.close()
            return

        # Move on.
        self._expect_frame = True

        # Finish up.
        self.connected = True
        self._connection.on_read = self._con_old_read
        self._safely_call(self.on_connect)

    ##### Control Methods #####################################################

    def close(self):
        """
        Close the WebSocket connection immediately.
        """
        if self._connection is None:
            return

        if hasattr(self, '_request'):
            del self._request
        if hasattr(self, '_headers'):
            del self._headers

        self.read_delimiter = None
        self._read_buffer = u""
        self._recv_buffer = ""

        self.connected = False
        self._update_addr()

        self._connection.close()
        self._connection = None

    def end(self, reason=1000):
        """
        Close the WebSocket connection nicely, waiting for any remaining data
        to be sent and sending a close frame before ending.

        =========  =========  ============
        Argument   Default    Description
        =========  =========  ============
        reason     ``1000``   *Optional.* The reason the socket is closing, from the ``CLOSE_REASONS`` dictionary.
        =========  =========  ============
        """
        if self._connection is None:
            return

        if self.version == 0:
            self._connection.end()
        else:
            # Look up the reason.
            rdesc = CLOSE_REASONS.get(reason, 'Unknown Close')
            reason = struct.pack("!H", reason) + rdesc

            self.write(reason, frame=FRAME_CLOSE)
            self._connection.end()

    ##### Public Event Handlers ###############################################

    def on_read(self, data):
        """
        Placeholder. Called when data is read from the WebSocket.

        =========  ============
        Argument   Description
        =========  ============
        data       A chunk of data received from the socket. Binary data will be provided as a string, and text data will be provided as a unicode string.
        =========  ============
        """
        pass

    def on_write(self):
        """
        Placeholder. Called after the WebSocket has finished writing data.
        """
        pass

    def on_connect(self):
        """
        Placeholder. Called after the WebSocket has connected to a client and
        completed its handshake.
        """

    def on_close(self):
        """
        Placeholder. Called after the WebSocket has finished closing.
        """
        pass

    def on_handshake(self, request, headers):
        """
        Placeholder. Called during the initial handshake, making it possible to
        validate the request with custom logic, such as Origin checking and
        other forms of authentication.

        If this function returns False, or a value that would be interpreted
        as False, the handshake will stop and the connection will be aborted.

        =========  ============
        Argument   Description
        =========  ============
        request    The :class:`~pants.contrib.http.HTTPRequest` being upgraded to a WebSocket.
        headers    An empty dict. Any values set here will be sent as headers when accepting (or rejecting) the connection.
        =========  ============
        """
        return True

    ##### I/O Methods #########################################################

    def write(self, data, binary=False, frame=None):
        """
        Write data to the WebSocket.

        ==========  ========  ============
        Arguments   Default   Description
        ==========  ========  ============
        data                  A string of data to write to the WebSocket. Unicode will be converted automatically.
        binary      False     If this is True, the data will be written as a binary frame, rather than text frame.
        ==========  ========  ============
        """
        if self._connection is None:
            log.warning("Attempted to write to closed %s #%d." %
                (self.__class__.__name__, self.fileno))
            return

        if not self.connected:
            log.warning("Attempted to write to disconnected %s #%d." %
                (self.__class__.__name__, self.fileno))
            return

        if binary and self.version == 0:
            raise ValueError("Attempted to send binary data to old-version WebSocket.")

        if isinstance(data, unicode):
            data = data.encode('utf-8')
        elif not isinstance(data, str):
            raise ValueError("Only strings may be written to WebSockets.")

        if not binary and '\xFF' in data:
            raise ValueError("Invalid character 0xFF in data to be sent.")

        if self.version == 0:
            self._connection.write("\x00%s\xFF" % data)
        else:
            if frame is None:
                if binary:
                    frame = 2
                else:
                    frame = 1

            self._connection.write(chr(0x80 | frame))
            if len(data) > 125:
                if len(data) > 65535:
                    self._connection.write(chr(127))
                    self._connection.write(struct.pack("!Q", len(data)))
                else:
                    self._connection.write(chr(126))
                    self._connection.write(struct.pack("!H", len(data)))
            else:
                self._connection.write(chr(len(data)))
            self._connection.write(data)

    ##### Internal Methods ####################################################

    def _safely_call(self, thing_to_call, *args, **kwargs):
        """
        Safely execute a callable.

        The callable is wrapped in a try block and executed. If an
        exception is raised it is logged.

        ==============  ============
        Argument        Description
        ==============  ============
        thing_to_call   The callable to execute.
        *args           The positional arguments to be passed to the callable.
        **kwargs        The keyword arguments to be passed to the callable.
        ==============  ============
        """
        try:
            return thing_to_call(*args, **kwargs)
        except Exception:
            log.exception("Exception raised on %s #%d." %
                    (self.__class__.__name__, self.fileno))

    def _update_addr(self):
        """
        Update the WebSocket's ``remote_addr`` and ``local_addr`` attributes
        with the values provided by the associated :class:`~pants.contrib.http.HTTPConnection`.
        """
        if self.connected and self._connection:
            self.remote_addr = self._connection.remote_addr
            self.local_addr = self._connection.local_addr
        else:
            self.remote_addr = None
            self.local_addr = None

    ##### Internal Event Handler Methods ######################################

    def _con_old_read(self, data):
        """
        Process incoming data, the old way.
        """
        self._recv_buffer += data

        if self._expect_frame:
            self._expect_frame = False
            self._frame = ord(self._recv_buffer[0])
            self._recv_buffer = self._recv_buffer[1:]

            if self._frame & 0x80 == 0x80:
                log.error("Unsupported frame type for old-style WebSockets %02X on %s #%d." %
                    (self._frame, self.__class__.__name__, self.fileno))
                self.close()
                return

        # Simple Frame.
        ind = self._recv_buffer.find('\xFF')
        if ind == -1:
            return

        # Read the data.
        self._read_buffer += self._recv_buffer[:ind].decode('utf-8')
        self._recv_buffer = self._recv_buffer[ind+1:]
        self._expect_frame = True

        # Act on the data.
        self._process_read_buffer()

    def _con_read(self, data):
        """
        Process incoming data.
        """
        self._recv_buffer += data

        if len(self._recv_buffer) < 2:
            return

        byte1 = ord(self._recv_buffer[0])
        fragment = 0x80 & byte1
        rsv1 = 0x40 & byte1
        rsv2 = 0x20 & byte1
        rsv3 = 0x10 & byte1
        opcode = 0x0F & byte1

        byte2 = ord(self._recv_buffer[1])
        mask = 0x80 & byte2
        length = 0x7F & byte2

        if length == 126:
            if len(self._recv_buffer) < 4:
                return
            length = struct.unpack("!H", self._recv_buffer[2:4])
            headlen = 4

        elif length == 127:
            if len(self._recv_buffer) < 10:
                return
            length = struct.unpack("!Q", self._recv_buffer[2:10])
            headlen = 10

        else:
            headlen = 2

        if mask:
            if len(self._recv_buffer) < headlen + 4:
                return
            mask = [ord(x) for x in self._recv_buffer[headlen:headlen+4]]
            headlen += 4

        if len(self._recv_buffer) < headlen + length:
            return

        # Got a full message!
        data = self._recv_buffer[headlen:headlen+length]
        self._recv_buffer = self._recv_buffer[headlen+length:]

        if mask:
            new_data = ""
            for i in xrange(len(data)):
                new_data += chr(ord(data[i]) ^ mask[i % 4])
            data = new_data
            del new_data

        # Control Frame Nonsense!
        if opcode == FRAME_CLOSE:
            self.end()
            return

        elif opcode == FRAME_PING:
            self.write(data, frame=FRAME_PONG)

        elif opcode == FRAME_BINARY and self.read_delimiter is EntireMessage:
            self._safely_call(self.on_read, data)
            return

        elif opcode == FRAME_TEXT:
            try:
                data = data.decode('utf-8')
            except UnicodeDecodeError:
                self.end(reason=1007)
                return

        else:
            data = data.decode('latin1')

        self._read_buffer += data
        self._process_read_buffer()

    def _con_close(self):
        """
        Close the WebSocket.
        """
        self.close()

    def _con_write(self):
        if self.connected:
            self._safely_call(self.on_write)

    ##### Internal Processing Methods #########################################

    def _process_read_buffer(self):
        """
        Process the read_buffer. This is only used when the ReadDelimiter isn't
        EntireMessage.
        """
        while self._read_buffer:
            delimiter = self.read_delimiter

            if delimiter is None or delimiter is EntireMessage:
                data = self._read_buffer
                self._read_buffer = u""
                self._safely_call(self.on_read, data)

            elif isinstance(delimiter, (int, long)):
                if len(self._read_buffer) < delimiter:
                    break
                data = self._read_buffer[:delimiter]
                self._read_buffer = self._read_buffer[delimiter:]
                self._safely_call(self.on_read, data)

            elif isinstance(delimiter, basestring):
                mark = self._read_buffer.find(delimiter)
                if mark == -1:
                    break
                data = self._read_buffer[:mark]
                self._read_buffer = self._read_buffer[mark + len(delimiter):]
                self._safely_call(self.on_read, data)

            else:
                log.warning("Invalid read_delimiter on %s #%d." %
                    (self.__class__.__name__, self.fileno))
                break

            if self._connection is None or not self.connected:
                break

###############################################################################
# Support Functions
###############################################################################

def challenge_response(headers, key3):
    """
    Calculate the response for a WebSocket security challenge and return it.
    """
    resp = hashlib.md5()

    for key in (headers.get('Sec-WebSocket-Key1'),
                headers.get('Sec-WebSocket-Key2')):
        n = ''
        s = 0
        for c in key:
            if c.isdigit(): n += c
            elif c == ' ': s += 1
        n = int(n)

        if n > 4294967295 or s == 0 or n % s != 0:
            raise ValueError("The provided keys aren't valid.")
        n /= s

        resp.update(
            chr(n >> 24 & 0xFF) +
            chr(n >> 16 & 0xFF) +
            chr(n >> 8  & 0xFF) +
            chr(n       & 0xFF)
        )

    resp.update(key3)
    return resp.digest()



###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

###############################################################################
# Imports
###############################################################################

import Cookie
import urllib
import urlparse
import zlib

from pants.contrib.http.utils import *

###############################################################################
# HTTPClient Class
###############################################################################

class HTTPClient(object):
    """
    An HTTP client, capable of communicating with most, if not all, servers
    using an incomplete implementation of HTTP protocol version 1.1.

    The behavior of an instance of HTTPClient is determined by that instance's
    :func:`on_response` function. That function may be changed by subclassing
    HTTPClient, assigning it directly, or supplying a suitable callable as the
    first argument when creating an instance of HTTPClient.

    =================  ========  ============
    Argument           Default   Description
    =================  ========  ============
    response_handler   None      *Optional.* A callable that will handle any received responses.
    max_redirects      ``5``     *Optional.* The number of times to follow a redirect issued by the server.
    keep_alive         True      *Optional.* Whether or not a single connection will be reused for multiple requests.
    unicode            True      *Optional.* Whether or not to attempt to convert the response body to unicode using the provided Content-Type header's encoding information.
    =================  ========  ============
    """
    def __init__(self, response_handler=None, max_redirects=5, keep_alive=True,
                unicode=True):
        if response_handler is not None:
            if not callable(response_handler):
                raise ValueError("response handler must be callable.")
            self.on_response = response_handler

        # Internal State
        self._stream = None
        self._processing = False
        self._requests = []
        self._server = None
        self._is_secure = False
        self._helper = None

        # External State
        self.keep_alive = keep_alive
        self.max_redirects = max_redirects
        self.unicode = unicode

    ##### General Methods #####################################################

    def get(self, url, timeout=30, headers=None, **kwargs):
        """
        Perform an HTTP GET request for the specified URL. Additional query
        parameters may be specified as keyword arguments. For example::

            client.get('http://www.google.com/search', q='test')

        Is equivalent to::

            client.get('http://www.google.com/search?q=test')

        =========  ========  ============
        Argument   Default   Description
        =========  ========  ============
        url                  The URL to request.
        timeout    ``30``    *Optional.* The time, in seconds, to wait for a response before erroring.
        headers    None      *Optional.* A dictionary of headers to send with the request. If none are provided, basic headers are set.
        =========  ========  ============
        """
        helper = self._helper
        if helper is None:
            helper = ClientHelper(self)

        if kwargs:
            query, fragment = urlparse.urlparse(url)[4:6]
            if query:
                query = "%s&%s" % (query, urllib.urlencode(kwargs, True))
                url = "%s?%s" % (url.partition('?')[0], query)
                if fragment:
                    url = "%s#%s" % (url, fragment)

            else:
                query = urllib.urlencode(kwargs, True)
                if fragment:
                    url = "%s?%s#%s" % (url.partition('#')[0], query, fragment)
                else:
                    url = "%s?%s" % (url, query)

        helper.requests.append(self._add_request('GET', url, headers, None,
                                timeout))

        return helper

    def post(self, url, timeout=None, headers=None, files=None, **kwargs):
        """
        Perform an HTTP POST request for the specified URL.

        =========  ========  ============
        Argument   Default   Description
        =========  ========  ============
        url                  The URL to request.
        timeout    ``30``    *Optional.* The time, in seconds, to wait for a response before erroring.
        headers    None      *Optional.* A dictionary of headers to send with the request. If none are provided, basic headers are set.
        files      None      *Optional.* A dictionary of files to send with the request. If this is provided, the dictionary keys should be equivalent to HTML form field names, and the values should be tuples of ``(filename, data)``.
        =========  ========  ============

        Any additional keyword arguments will be sent in the request body as
        POST variables.
        """
        helper = self._helper
        if helper is None:
            helper = ClientHelper(self)

        body = ''

        if headers is None:
            headers = {}

        if headers.get('Content-Type', '') == 'application/x-www-form-urlencoded' and files:
            raise ValueError("Cannot send files with Content-Type "
                             "'application/x-www-form-urlencoded'.")

        if files:
            headers['Content-Type'] = 'multipart/form-data'
        elif not 'Content-Type' in headers:
            headers['Content-Type'] = 'application/x-www-form-urlencoded'

        if headers['Content-Type'] == 'multipart/form-data':
            boundary, body = encode_multipart(kwargs, files)

            headers['Content-Type'] = 'multipart/form-data; boundary=%s' % \
                boundary

        elif kwargs:
            body = urllib.urlencode(kwargs, True)

        helper.requests.append(self._add_request('POST', url, headers, body,
                                timeout))

        return helper

    def process(self):
        """
        Useful for testing, and other synchronous work, this function will
        start the Pants engine and block until all outstanding requests
        complete.
        """
        if not self._requests:
            return

        self._processing = True
        Engine.instance().start()

    ##### Public Event Handlers ###############################################

    def on_response(self, response):
        """
        Placeholder. Called when an HTTP response is received.

        =========  ============
        Argument   Description
        =========  ============
        response   The received HTTP response.
        =========  ============
        """
        pass

    ##### Private Methods #####################################################

    def _add_request(self, method, url, headers, body, timeout, append=True):
        u = url.lower()
        if not (u.startswith('http://') or u.startswith('https://')):
            raise ValueError("Can only make HTTP or HTTPS requests with HTTPClient.")

        parts = urlparse.urlparse(url)

        # Build our headers.
        if headers is None:
            headers = {}

        if not 'Accept-Encoding' in headers:
            headers['Accept-Encoding'] = 'deflate, gzip'

        if not 'Host' in headers:
            headers['Host'] = parts.netloc

        if not 'User-Agent' in headers:
            headers['User-Agent'] = USER_AGENT

        if body:
            headers['Content-Length'] = len(body)

        path = parts.path or '/'
        if parts.query:
            path = '%s?%s' % (path, parts.query)
        if parts.fragment:
            path = '%s#%s' % (path, parts.fragment)

        request = [method, url, parts, headers, body, timeout, None, time(),
                    path, 0]

        if append:
            self._requests.append(request)

            # If we're just starting, start to process.
            if len(self._requests) == 1:
                callback(self._process_request)

        return request

    def _process_request(self):
        """
        Starts processing the first request on the stack.
        """
        if not self._requests:
            if self._stream:
                self._stream.close()
                self._stream = None
            if self._processing:
                self._processing = False
                Engine.instance().stop()
            return

        request = self._requests[0]

        request.append(
            Engine.instance().defer(request[5], self._request_timeout, request))

        port = request[2].port
        if not port:
            if request[2].scheme.lower() == 'https':
                port = 443
            else:
                port = 80

        host = "%s:%d" % (request[2].hostname, port)

        if self._stream:
            if not self._server == host.lower() or not \
                    self._is_secure == (request[2].scheme.lower() == 'https'):
                self._stream.end()
                return

        if not self._stream:
            # Store the current server.
            self._server = host.lower()

            # Create a Stream, hook into it, and connect.
            self._stream = Stream()

            self._stream.on_close = self._on_close
            self._stream.on_connect = self._on_connect

            self._is_secure = request[2].scheme.lower() == 'https'
            if self._is_secure:
                raise Exception("SSL has not yet been implemented in this version of Pants.")
                self._stream.startTLS()

            self._stream.connect((request[2].hostname, port))
            return

        # If we got here, we're connected, and to the right server. Do stuff.
        self._stream.write('%s %s HTTP/1.1%s' % (request[0], request[8], CRLF))
        for k, v in request[3].iteritems():
            self._stream.write('%s: %s%s' % (k, v, CRLF))

        if request[4]:
            self._stream.write('%s%s' % (CRLF, request[4]))
        else:
            self._stream.write(CRLF)

        # Now, wait for a response.
        self._stream.on_read = self._read_headers
        self._stream.read_delimiter = DOUBLE_CRLF

    ##### Internal Event Handlers #############################################

    def _on_connect(self):
        if self._requests:
            self._process_request()
        else:
            self._stream.end()

    def _on_close(self):
        """
        In the event that the connection is closed, see if there's another
        request to process. If so, reconnect to the given host.
        """
        self._stream = None
        self._is_secure = False
        self._process_request()

    def _on_response(self):
        """
        Call the response handler.
        """
        request = self._requests.pop(0)
        try:
            request[-1].cancel()
            left = request[-1].end - Engine.instance().time
        except Exception:
            left = request[5]
            pass

        response = self.current_response

        close_after = response.headers.get('Connection', '') == 'close'
        close_after &= self.keep_alive

        # Is this a 100 Continue?
        if response.status == 100:
            self.current_response = None
            del response

            # Process the request.
            if close_after:
                if self._stream:
                    self._stream.close()
                    return

            self._process_request()
            return

        # Did we catch a redirect?
        if response.status in (301,302) and request[9] <= self.max_redirects:
            # Generate a new request, using the new URL.
            new_url = urlparse.urljoin(response.full_url,
                        response.headers['Location'])

            new_headers = request[3].copy()
            del new_headers['Host']

            new_req = self._add_request(request[0], new_url, new_headers,
                                        request[4], left, False)
            new_req[6] = request[6]
            new_req[7] = request[7]
            new_req[9] = request[9] + 1

            new_req.append(
                Engine.instance().defer(left, self._request_timeout, new_req))

            self._requests.insert(0, new_req)
            self.current_response = None
            del response

            # Process the request.
            if close_after:
                if self._stream:
                    self._stream.close()
                    return

            self._process_request()
            return

        # Try converting to unicode?
        if self.unicode:
            content_type = response.headers.get('Content-Type','')
            if 'charset=' in content_type:
                content_type, _, encoding = content_type.partition('charset=')
                try:
                    response.body = response.body.decode(encoding)
                except (LookupError, UnicodeDecodeError):
                    pass

        # Determine the handler function to use.
        if callable(request[6]):
            func = request[6]
        else:
            func = self.on_response

        # Call the handler function.
        try:
            func(0, response)
        except Exception:
            log.exception('Error in HTTP response handler.')

        # Process the next request.
        self.current_response = None

        if close_after:
            if self._stream:
                self._stream.close()
                return

        self._process_request()

    def _request_timeout(self, request):
        if not request in self._requests:
            return

        self._requests.remove(request)

        if callable(request[6]):
            func = request[6]
        else:
            func = self.on_response

        try:
            func(1, None)
        except Exception:
            log.exception('Error in HTTP response handler.')

    def _read_body(self, data):
        """
        Read the response body, decompress it if necessary, and then call the
        response handler.
        """
        resp = self.current_response
        if resp._decompressor:
            resp.body = resp._decompressor.decompress(data)
            resp.body += resp._decompressor.flush()
            del resp._decompressor
        else:
            resp.body = data
        self._on_response()

    def _read_additional_headers(self, data):
        resp = self.current_response

        if data:
            resp._additional_headers += '%s%s' % (data, CRLF)
            return

        headers = read_headers(resp._additional_headers)
        del resp._additional_headers

        # Did we get an additional header for Content-Encoding?
        enc = resp.headers.get('Content-Encoding', '')

        for k,v in headers.iteritems():
            if k in resp.headers:
                if not isinstance(resp.headers[k], list):
                    resp.headers[k] = [resp.headers[k]]
                if isinstance(v, list):
                    resp.headers[k].extend(v)
                else:
                    resp.headers[k].append(v)
            else:
                resp.headers[k] = v

        new_enc = resp.headers.get('Content-Encoding', '')
        if (new_enc == 'gzip' or new_enc == 'deflate') and enc == '':
            if new_enc == 'gzip':
                resp.body = zlib.decompress(resp.body, 16 + zlib.MAX_WBITS)
            elif new_enc == 'deflate':
                resp.body = zlib.decompress(resp.body, -zlib.MAX_WBITS)

        # Finally, handle it.
        self._on_response()

    def _read_chunk_head(self, data):
        """
        Read a chunk header.
        """
        if ';' in data:
            data, ext = data.split(';', 1)
        else:
            ext = ''

        length = int(data.strip(), 16)

        if length == 0:
            resp = self.current_response
            if resp._decompressor:
                resp.body += resp._decompressor.flush()
                del resp._decompressor

            self._stream.on_read = self._read_additional_headers
            resp._additional_headers = ''
            self._stream.read_delimiter = CRLF

        else:
            self._stream.on_read = self._read_chunk_body
            self._stream.read_delimiter = length + 2

    def _read_chunk_body(self, data):
        """
        Read a chunk body.
        """
        resp = self.current_response

        if resp._decompressor:
            resp.body += resp._decompressor.decompress(data[:-2])
        else:
            resp.body += data[:-2]

        self._stream.on_read = self._read_chunk_head
        self._stream.read_delimiter = CRLF

    def _read_headers(self, data):
        """
        Read the headers of an HTTP response from the socket, and the response
        body as well, into a new HTTPResponse instance. Then call the request
        handler.
        """
        do_close = False

        try:
            initial_line, data = data.split(CRLF, 1)
            try:
                try:
                    http_version, status, status_text = initial_line.split(' ', 2)
                    status = int(status)
                except ValueError:
                    http_version, status = initial_line.split(' ')
                    status = int(status)
                    status_text = HTTP.get(status, '')
            except ValueError:
                raise BadRequest('Invalid HTTP status line %r.' % initial_line)

            # Parse the headers.
            headers = read_headers(data)

            # Construct an HTTPResponse object.
            self.current_response = response = HTTPResponse(self,
                self._requests[0], http_version, status, status_text, headers)

            # Do we have a Content-Encoding header?
            if 'Content-Encoding' in headers:
                encoding = headers['Content-Encoding']
                if encoding == 'gzip':
                    response._decompressor = zlib.decompressobj(16+zlib.MAX_WBITS)
                elif encoding == 'deflate':
                    response._decompressor = zlib.decompressobj(-zlib.MAX_WBITS)

            # Do we have a Content-Length header?
            if 'Content-Length' in headers:
                self._stream.on_read = self._read_body
                self._stream.read_delimiter = int(headers['Content-Length'])

            elif 'Transfer-Encoding' in headers:
                if headers['Transfer-Encoding'] == 'chunked':
                    self._stream.on_read = self._read_chunk_head
                    self._stream.read_delimiter = CRLF
                else:
                    raise BadRequest("Unsupported Transfer-Encoding: %s" % headers['Transfer-Encoding'])

            # Is this a HEAD request? If so, then handle the request NOW.
            if response.method == 'HEAD':
                self._on_response()

        except BadRequest, e:
            log.info('Bad response from %r: %s',
                self._server, e)
            do_close = True

        except Exception:
            log.exception('Error handling HTTP response.')
            do_close = True

        # Clear the way for the next request.
        if do_close:
            self._requests.pop(0)
            self.current_response = None
            if self._stream:
                self._stream.close()
                self._stream = None

class ClientHelper(object):
    """
    An instance of this class is returned by calls to :func:`HTTPClient.get`
    and :func:`HTTPClient.post` to allow for a bit of decorator magic.

    For further information, please see the :class:`HTTPClient` documentation.
    """
    def __init__(self, parent):
        self.parent = parent
        self.requests = []
        self._responses = []

    def get(self, *a, **kw):
        self.parent._helper = self
        try:
            return self.parent.get(*a, **kw)
        finally:
            self.parent._helper = None

    def post(self, *a, **kw):
        self.parent._helper = self
        try:
            return self.parent.post(*a, **kw)
        finally:
            self.parent._helper = None

    def _collect(self, status, response):
        self._responses.append((status, response))

    def fetch(self):
        """
        Perform all the pending requests and return them. If there is more than
        one outstanding request, the results will be returned as a list of
        :class:`HTTPResponse` instances. Otherwise, a single HTTPResponse
        instance will be returned.
        """
        for req in self.requests:
            req[6] = self._collect

        self.process()

        out = self._responses
        self._responses = []

        if len(out) == 1:
            return out[0]
        return out

    def process(self):
        return self.parent.process()

    def __call__(self, func):
        for req in self.requests:
            req[6] = func
        return func

###############################################################################
# HTTPResponse Class
###############################################################################

class HTTPResponse(object):
    """
    Instances of this class represent singular HTTP responses that an instance
    of :class:`HTTPClient` has received. Such instances contain all the
    information needed to act upon a response.

    This class should, generally, not be used directly. Instead, allow the
    HTTPClient to create instances for you.

    =============  ============
    Argument       Description
    =============  ============
    client         The instance of :class:`HTTPClient` that received this response.
    request        The request list, containing all the information passed to the call that generated the request responsible for this response.
    http_version   The HTTP protocol version used in this response. This will almost always be one of: ``HTTP/1.0`` or ``HTTP/1.1``.
    status         The `HTTP status code <http://en.wikipedia.org/wiki/Http_status_codes>`_ received with this response.
    status_text    The human readable status message that goes with the received status code.
    headers        A dictionary of HTTP headers received with this response.
    =============  ============
    """

    def __init__(self, client, request, http_version, status, status_text,
                    headers):
        self.body = ''
        self.client = client
        self.headers = headers
        self.method = request[0]
        self.protocol = request[2].scheme
        self.request = request
        self.uri = request[8]
        self.version = http_version

        self.host = client._server
        if self.host.endswith(':80'):
            self.host = self.host[:-3]

        self.status = status
        self.status_text = status_text

        self._decompressor = None

        # Timing Information
        self._start = request[7]
        self._finish = time()
        self.time = self._finish - self._start

    ##### Properties ##########################################################

    @property
    def cookies(self):
        """
        An instance of :class:`Cookie.SimpleCookie` representing the cookies
        received with this response.
        """
        try:
            return self._cookies
        except AttributeError:
            self._cookies = cookies = Cookie.SimpleCookie()
            if 'Set-Cookie' in self.headers:
                raw = self.headers['Set-Cookie']
                if isinstance(raw, list):
                    for i in raw:
                        cookies.load(i)
                else:
                    cookies.load(raw)
            return self._cookies

    @property
    def full_url(self):
        """
        The full URL of the request that generated this response.
        """
        return "%s://%s%s" % (self.protocol, self.host, self.uri)




###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

###############################################################################
# Imports
###############################################################################

import cStringIO
import re
import sys
import traceback

from http import log
from web import error

###############################################################################
# WSGIConnector Class
###############################################################################

class WSGIConnector(object):
    """
    This class acts as a request handler for :class:`pants.contrib.http.HTTPServer`
    and provides a simple interface for hosting `WSGI <http://en.wikipedia.org/wiki/WSGI>`_
    compatible applications.

    When called, it constructs a proper environment for the WSGI call, and
    calls it within a try block to catch errors. A ``500 Internal Server Error``
    page is displayed if an exception bubbles up from the WSGI application.

    ============  ============
    Argument      Description
    ============  ============
    application   The WSGI application that will handle incoming requests.
    debug         *Optional.* Whether or not to display tracebacks and additional information about a request within the output 500 Internal Server Error pages.
    ============  ============
    """
    def __init__(self, application, debug=False):
        self.app = application
        self.debug = debug

    def attach(self, application, path, domain=None):
        """
        Attach the WSGIConnector to an instance of
        :class:`pants.contrib.web.Application` at the given route.

        ============  ========  ============
        Argument      Default   Description
        ============  ========  ============
        application             The :class:`~pants.contrib.web.Application` to attach to.
        path                    The path to serve requests from.
        domain        None      *Optional.* The domain to serve requests upon.
        ============  ========  ============
        """
        path = re.compile("^%s(.*)$" % re.escape(path))
        application._insert_route(
            path, self, domain, "WSGIConnector", ['HEAD','GET','POST','PUT'],
            None, None)

    def __call__(self, request):
        """
        Handle the given request.
        """
        # Make sure this plays nice with Web.
        request.auto_finish = False

        def start_response(status, head):
            request.send_status(status)
            if isinstance(head, list):
                head = dict(head)
            request.send_headers(head)

            return request.write

        # Build an environment for the WSGI application.
        environ = {
            'REQUEST_METHOD'    : request.method,
            'SCRIPT_NAME'       : '',
            'PATH_INFO'         : request.path,
            'QUERY_STRING'      : request.query,
            'SERVER_NAME'       : request.headers.get('Host','127.0.0.1'),
            'SERVER_PROTOCOL'   : request.version,
            'REMOTE_ADDR'       : request.remote_ip,
            'GATEWAY_INTERFACE' : 'WSGI/1.0',
            'wsgi.version'      : (1,0),
            'wsgi.url_scheme'   : request.protocol,
            'wsgi.input'        : cStringIO.StringIO(request.body),
            'wsgi.errors'       : sys.stderr,
            'wsgi.multithread'  : False,
            'wsgi.multiprocess' : False,
            'wsgi.run_once'     : False
        }

        if isinstance(request.connection.server.local_addr, tuple):
            environ['SERVER_PORT'] = request.connection.server.local_addr[1]

        if hasattr(request, 'arguments'):
            environ['wsgiorg.routing_args'] = (request.arguments, {})
        elif hasattr(request, 'match'):
            environ['wsgiorg.routing_args'] = (request.match.groups(), {})

        if 'Content-Type' in request.headers:
            environ['CONTENT_TYPE'] = request.headers['Content-Type']
        if 'Content-Length' in request.headers:
            environ['CONTENT_LENGTH'] = request.headers['Content-Length']

        for k,v in request.headers.iteritems():
            environ['HTTP_%s' % k.replace('-','_').upper()] = v

        # Run the WSGI Application.
        try:
            result = self.app(environ, start_response)
        except Exception, e:
            log.exception('Exception running WSGI application for: %s %s',
                request.method, request.path)

            if not self.debug:
                body, status, headers = error(500, request=request, debug=False)
            else:
                resp = u''.join([
                    u"<h2>Traceback</h2>\n",
                    u"<pre>%s</pre>\n" % traceback.format_exc(),
                    u"<h2>HTTP Request</h2>\n",
                    request.__html__(),
                    ])
                body, status, headers = error(resp, 500, request=request,
                    debug=True)

            request.send_status(500)

            if not 'Content-Length' in headers:
                headers['Content-Length'] = len(body)

            request.send_headers(headers)
            request.write(body)
            request.finish()
            return

        # Finish up anything in result.
        if result:
            try:
                for thing in result:
                    request.write(thing)
            finally:
                if hasattr(result, 'close'):
                    result.close()
                del result

        request.finish()



###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

###############################################################################
# Imports
###############################################################################

import base64
import inspect
import logging
import mimetypes
import os
import re
import time
import traceback
import urllib

from datetime import datetime, timedelta
from pants import __version__ as pants_version
from http import CRLF, HTTP, HTTPServer, HTTPRequest, SERVER, SERVER_URL, date

try:
    import simplejson as json
except ImportError:
    import json

__all__ = ('Application', 'HTTPException', 'HTTPTransparentRedirect', 'abort',
    'all_or_404', 'error', 'redirect', 'url_for', 'HTTPServer', 'FileServer')

###############################################################################
# Cross Platform Hidden File Detection
###############################################################################

def _is_hidden(file, path):
    return file.startswith(u'.')

if os.name == 'nt':
    try:
        import win32api, win32con
        def _is_hidden(file, path):
            if file.startswith(u'.'):
                return True
            file = os.path.join(path, file)
            try:
                if win32api.GetFileAttributes(file) & win32con.FILE_ATTRIBUTE_HIDDEN:
                    return True
            except Exception:
                return True
            return False
    except ImportError:
        pass

###############################################################################
# Logging
###############################################################################

log = logging.getLogger(__name__)

###############################################################################
# Constants
###############################################################################

HAIKUS = {
    400: u'Something you entered<br>'
         u'transcended parameters.<br>'
         u'So much is unknown.',

    401: u'To access this page,<br>'
         u'one must know oneself; but then:<br>'
         u'inform the server.',

    403: u'Unfortunately,<br>'
         u'permissions insufficient.<br>'
         u'This, you cannot see.',

    404: u'You step in the stream,<br>'
         u'But the water has moved on.<br>'
         u'This page is not here.',

    410: u'A file that big?<br>'
         u'It might be very useful.<br>'
         u'But now it is Gone.',

    413: u'Out of memory.<br>'
         u'We wish to hold the whole sky,<br>'
         u'But we never will.',

    418: u'You requested coffee,<br>'
         u'it is neither short nor stout.<br>'
         u'I am a teapot.',

    500: u'Chaos reigns within.<br>'
         u'Reflect, repent, and reboot.<br>'
         u'Order shall return.'
}

if os.name == 'nt':
    HAIKUS[500] = (u'Yesterday it worked.<br>'
        u'Today, it is not working.<br>'
        u'Windows is like that.')

HTTP_MESSAGES = {
    401: u'You must sign in to access this page.',
    403: u'You do not have permission to view this page.',
    404: u'The page at <code>%(uri)s</code> cannot be found.',
    500: u'The server encountered an internal error and cannot display '
         u'this page.'
}

IMAGES = {
    'audio'     : u"iVBORw0KGgoAAAANSUhEUgAAABIAAAAQCAYAAAAbBi9cAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAZpJREFUeNqsU01Lw0AQbdI0aVIVUiSlWKiV/gARiveC2ptehf4Rz+JJ0aM38ebJgzfvvQiWUhooqAeLYCkRUYRa8tHEN7IpIRS3ogsvuzs782Yz81YIgiDxH0P6ha/QbrdLvu/XgC1cIFWpVLZhd7lEpmlmXdetImgTwRuYl2Muc8DbT0SpZrN5huD65DqCkBBFkTAE3pFgCWaNR5RB9joFS5L0Ksvyg6qq97qum0CPJTrHLPJqlKFPPp8/KRQKDSw/IvgEsqw2AY/oOxNILjE9sWCbwSOCINZuXtf6wDPg81oqcs69WUhmIfq7IGMlEFut1u54PN6HvYROXpMiphEJnU5n1bbtUziuwbER41VBcowzgzZY1yANZ9qvKSC5gOM6acTzvCppKDI00hLZQruiKDfR+oVEmWQyqYWOBOz7EZ14xWLxMJ1Od6FqV9O023K5fAD7aKJ8VovFwWCwY1nWnuM4K8h2l8vljgzDuMLZCyCTPoESsMCexSNgAU6USAXo7dCjnGcK7jEdjVhhZaZ4mQlzGJLQ+BJgAITfplvWq5n7AAAAAElFTkSuQmCC",
    'document'  : u"iVBORw0KGgoAAAANSUhEUgAAABIAAAAQCAYAAAAbBi9cAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAS1JREFUeNpi/P//PwM1AAsan+306dPngYZr4dPEyMh4zdTU1BDI/IXLIF4g1mJiYgIpBgsguxjEhvK1oGrf4jKIE0QoKCjUi4iI3AaxX79+rfXw4cMaqEvAGGoYJz6vgZ1x//79RiCGuwpmCFp4MuIzCAykpaXLpaSkjkNdZAh00URSAxts65MnTzqBGEUcFG4kGQQLB2RvoVtElEEgoKKiUiEgIPAIpA/dnhcvXug/fvy4nCiDbt++3UFpggQDCQmJBllZ2X1A5j80KeZnz55ZP336tI0og4DObwBhil0kIyNTJikpeRLI/IsmxfTy5UvjR48e9RMV/cDA7AJiksIIPXH8Y2dnvwBKM/gwSA16+DGipQshINYAYilc3gaCP0D8DIhvAPE7mCBAgAEAx0h2pQytmCsAAAAASUVORK5CYII=",
    'folder'    : u"iVBORw0KGgoAAAANSUhEUgAAABIAAAAKCAYAAAC5Sw6hAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAQ9JREFUeNqsUT2LwkAQvd3EaDDIESxiaQqLgCDk59z9Gu9XWN8/sUzE0i4iNloEES3EZJP1jWRExK/Chbezszvzdt6M0Fp/fWKZN349iqItbIMcwzD+wjAc4qheEUnakiRpxXH8n6ZpVwjRAM6PqPYHxn63IlkURR9Jv5ZljeiSicqy9FHh7kn+1nGcQRAES6pI+L6/llIeIKXJJPBJ2hl0vgXFAd+e51ExgrbSNM0MCbM8z+v3vmXySu7lDti4rkv901QRvRxBNFVKCQ58Z5rIWVWD0Dy1I/ozQbJiEvrxERnfQ8kSJr8ef4amjaG9A2QItK7lPFq2bcdMxFKIsAa0gV5lXzHtgTmwAA4nAQYAHA9ij4jhqJgAAAAASUVORK5CYII=",
    'icon'      : u"iVBORw0KGgoAAAANSUhEUgAAACgAAAAoCAYAAACM/rhtAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAQ5JREFUeNpiYWBg+M8wiAELiGhI9RuUjmuYvYmBiWGQA8YhEcVT995EETx96TrDgsIAhoT+DUQbBFJvnNEL5+uqK5OkF2SXqZ4mini2szrEgeiOQwaXb94ly+fE6kP2CMhudEeyYHMUqaFHKQDZBbMT3S1Y0yDMcaSE3tkZxShRTAqAhSLIkVjTILbQIjdqyU0OIEeiuwPkYJaBcBAxaRYWqoO+HByZDgRlGKoW1NR2GLm5maYOpKajhlQapEoI4qp3qVF0US2K0WsBalWVVI3i////M4LwaDk46sBRB446cNSBow4cdeCoA0cdOOrAUQeOOpDUThMpI6KU9vZIAVQdo4Z1mBgZGalmJkCAAQB+2V2B4VtJPwAAAABJRU5ErkJggg==",
    'image'     : u"iVBORw0KGgoAAAANSUhEUgAAABUAAAATCAYAAAB/TkaLAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAphJREFUeNqsVE1rE1EU7UxmEjMl0QYxIU4SRi2KH9CkBhWKC3FbRfAH+AfcuBH/hZsiCC66U1y4URcKFnQjk5As1G6UhsQmJCkGZpIxyeRjPDe8J9OhaSL44DBvHu/dc9+55z7BcZyF/z0kQM7n80/H4/E9WhAEYd8GTor1r9lsdhVTe56gi8DdWCz2IJlM5jEfe/YItVrtYrVafVKv11Xs25kVVASCdDASiXzBdxeoeLAbj8eL+FqdTudCoVBI0/5ZmfpwRWcwGOyxICNvpqQCyWLb9rXhcPgol8tVRVF8ibU3mUzmg/d2kks7CnZQ1RxOlEqlPrZarQoyvtzr9dZAcB8ELQR/C5LnIHhHBNK8FaXbjEajdiKR2MIvaX+k3+8r0PtSu92+CpLXxWLxYTqdfiz9i1UoKJNIpqCBQEDRNI3q8BkBz/l8vmXMAyI/0Gw2M1MKIJRKpTNIVAFIuz5g0hHgJ1CyLMsEoRYOh3UqPs9UME1zU9f1zRnJeklJa7tcLl8hdVRVpaADiRl73+ZpDTDlJgG44o7f79clSTIoqDiPlkRCaDQaN9F9z6DfDebxBbhhCXa8HgwGP+H3N6/+SFGU991udx0Zid4s+ZBleRu2OUHtTICVKijMC+zXSO9oNEqu6E2SwMJRfKlqKlXU3W2wywpsEyU7IVDXMIx1FOQkEbsfIhB+g5VuMWcMKVML+A7UqLtcQRfR7xs4fOwgKdyBcXWdxRnyjqKJweAex7F5C6a+zfWbNmAlatXuX+JD3tOJLGjF0/DwKrx4HgRnUelTpD13BXT9hfZcw+8OfxYP6yi6zg/YZA+v1DYlRICmS3DBCpGguMuhUOgVu+Vgnkzdz6Of/MigACFGIrHOqrAkJuOPAAMATZ5MP7rfmUUAAAAASUVORK5CYII=",
    'pants'     : u"iVBORw0KGgoAAAANSUhEUgAAAKAAAACgCAYAAACLz2ctAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAhxJREFUeNrs3U1OwkAYgGEHJyndsSKkm3I9E6/jNbgLN+gJum9Cp8q2CsbYH1qeZwmE6PBm5kuLMZxOp/cQwgtM7XA4pLjb7br9fp8sB1PLsixsLANzEiCziv0HjsejVWE0VVXZAXEEgwARIAIEASJAmN6364D96zRgB0SAIEDWPwMOfS+4P1Mu7V7z0n/+R/t93QvGEQwCRIAgQASIAGEWceg3fLbrZmvT/7zG/jztgDiCESA8xgz43+8Dmvmeaya0A+IIBgGyjhnQDMeYM77vA+IIBgEiQBAgAkSAIEAECAJEgCBABAgCRIAwimgJhuXvou2ACBAEiBlw/TNen5nPDogAQYCYAZfPTGcHRIAgQAQIAkSAIEAECAJEgCBAFsi94F/4vyl2QAQIjuDZbbfbEGMMVz893zRNulwuXUrJYglwhONiswllWX7ce835fH77itRiOYKH17Ztd+/568ZY17U1FeD0bpzKOIJFZwcEASJAECACBAEiQARoCRAgAgQBIkAQIAIEASJAECACBAEiQBAgAgQBIkAQIAIEASJAECACBAEiQBAgAgQBIkAQIAIEASJABAgCRIAgQAQIAkSAIEAECAJEgCBABAgCRIAgQAQIAkSAIEAECAJEgCBABAgCRIAgQAQIAkSAIEAEiABBgAgQBIgAQYAIEMYV+w9UVWVVBmQ97YAIEG4cwUVRpDzPXy3FMKzn33wKMACVd1AkmFTspgAAAABJRU5ErkJggg==",
    'video'     : u"iVBORw0KGgoAAAANSUhEUgAAABIAAAAPCAYAAADphp8SAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAe9JREFUeNqsU0tLAlEUdmZ8Tki5qHxvtV3gLgmC6C+06me06CG5KFBo0b6lEEI7qWVB5CJCXKuLFsngeyEjk4yOY98d7pWrGW0aOJxz73z3O+d+51xhOp3a/uOzUy+Wy+VDj8dTHI/H27quPzKAIAiW5xO6XK59OMUwjK1EIvGA2GRE8mQyuYU/crvd66PRyOb3+48jkcg74WDJ6vV6st1uZ2RZ3kCyTeCusf8E0yyibrcbRGavpmkFlhUkb3BNmMGqj0ajRRDZ+v1+nrvVCiMSJElykp1QKJQ2TdPVbDbPKMknKZtVxE4SHG7gbLVa51g6mUYoRhCJBtBHB5FJ9TA4EhuNjSU4gRfbErXT6WTZulqt7sXj8Tu+okqlckAChmONmCMiHwTOiqJoNBqN1GAwyJVKpdxim8lhgoM3IEHqBxEpE+3UoJdO1sFgMA0tXhBOKETCwR1FUTIMx4/E3NV6vd4lJ+gzXIPvWiAQeAXRDPfr1cLh8AWyyejGCSVZ2rUF3NKrDWDCX11bwM2Ipj6fT4XIQ2S4YT9qtVoyFovd8xWhk7skYDjo1GKTb6fBCGNf8Hq9Hw6HYwgdrlRVzaNr+cWuISGZ+lM8ERnzZJ219KLlrZJGwdx0UtdoJUP+rcE8dAD7sC/yNGBt4r8FGADC3BrRMDVuEAAAAABJRU5ErkJggg==",
    'zip'       : u"iVBORw0KGgoAAAANSUhEUgAAABIAAAAQCAYAAAAbBi9cAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAVdJREFUeNpi/P//PwM1AAsan+306dPngYZrIQtyc3Pr/fnzR+Lnz5+7QHxGRsZrpqamhkDmL1wG8QKxFlQxg5ycXMbfv3+/Pn369BJMDOoDLajat7gM4oRpAIHHjx/PANGysrKZQAMEnzx50gaTg6nFZRBcFUiDoqJi7u/fv/8ADZiO5iIUtdgMYkB20f379yeDaAUFhYR///6JPHr0qAfNMPwGgRSCNMjLyxcCDfj64MGDBaTGGgp4+PBhP4iWkZHJBFLgMMKllgmfQVJSUrUmJiZ2P378YAUGfBvQdQzkGrQPSD0ChtEekFeZmJjIMwgI3p06derSx48feQmFEU6DQN64c+eOvZmZmcHr169N8XmLYGC/e/duBtBFGDFKkkGwtISUkhnwZXB0r/1jZ2e/BNIMC1wYGxmD1IDUoliMZosQEGuAIgyPa/8A8TMgvgHyPUwQIMAA22WMeFl8he8AAAAASUVORK5CYII="
}

IMAGES['icon'] = base64.b64decode(IMAGES['icon'])

PAGE_CSS = u"""html, body { margin: 0; padding: 0; min-height: 100%%; }
body {
	font-family: Calibri,"Arial","Helvetica",sans-serif;
	background: #EEE;
	background-image: -webkit-gradient( linear, left bottom, left top,
		color-stop(0, #ccc), color-stop(0.5, #eee) );
    background-image: -moz-linear-gradient( center bottom, #ccc 0%%, #eee 50%% );
}

table.dir td,a { color: #666; }
h1, a:hover { color: #444; }

a { text-decoration: none; }
a:hover { text-decoration: underline; }

div.document,.left,pre,table.dir th:first-child,table.dir td:first-child {
    text-align: left; }
.thingy,.center,.footer { text-align: center; }
table.dir td,table.dir th,.right { text-align: right; }

table.dir td,table.dir th,.thingy > h1 { margin: 0; }
p { margin-bottom: 0; }
table.dir a,pre { display: block; }
pre {
	background: #ddd;
    background-color: rgba(199,199,199,0.5);
	text-align: left;
	border-radius: 5px;
	-moz-border-radius: 5px;
	padding: 5px;
}

table.dir { width:100%%; border-spacing: 0; }
table.dir td,table.dir th { padding: 2px 5px; }
table.dir td { border-top: 1px solid transparent; border-bottom: 1px solid transparent; }
table.dir tr:first-child td { border-top: none; }
table.dir tr:hover td { border-color: #ccc; }
table.dir td.noborder { border-color: transparent !important; }
table.dir th { border-bottom: 1px solid #ccc; }

.footer,.faint { color: #aaa; }
.footer .debug { font-size: 0.9em; font-family: Consolas,monospace; }
.haiku { margin-top: 20px; }
.haiku + p { color: #777; }
.spacer { padding-top: 60px; }
.column { max-width: 960px; min-width: 600px; margin: 0px auto; }
.footer { padding-top: 10px; }

a.icon { padding-left: 23px; background-position: left; }
a.icon,.thingy { background-repeat: no-repeat; }

a.folder { background-image: url("data:image/png;base64,%s"); }
a.document { background-image: url("data:image/png;base64,%s"); }
a.image { background-image: url("data:image/png;base64,%s"); }
a.zip { background-image: url("data:image/png;base64,%s"); }
a.audio { background-image: url("data:image/png;base64,%s"); }
a.video { background-image: url("data:image/png;base64,%s"); }

.thingy { background-color: #FFF; background-position: center; color: #000;
	border: 5px #ddd solid;
	-moz-border-radius: 25px;
	border-radius: 25px;
	padding: 50px;
	margin: 0 50px;
}""" % (IMAGES['folder'], IMAGES['document'], IMAGES['image'], IMAGES['zip'],
    IMAGES['audio'], IMAGES['video'])
PAGE_CSS = PAGE_CSS.replace('%','%%%%')

PAGE = u"""<!DOCTYPE html>
<html><head><title>%%s</title><style>%s</style></head><body>
<div class="column"><div class="spacer"></div><div class="thingy">
%%s
</div><div class="footer"><i><a href="%s">%s</a><br>%%%%s</i>
<div class="debug">%%%%s</div></div>
<div class="spacer"></div></div></body></html>""".replace('\n','') % (
    PAGE_CSS, SERVER_URL, SERVER)

DIRECTORY_PAGE = PAGE % (
    u'Index of %s',
    u"""<h1>Index of %s</h1>%s<table class="dir"><thead><tr>
<th style="width:50%%">Name</th><th>Size</th>
<th class="center" colspan="2">Last Modified</th></tr></thead>%s
</table>"""
    )

ERROR_PAGE = PAGE % (
    u'%d %s',
    u'<h1>%d<br>%s</h1>%s%s'
    )

# Regular expressions used for various types.
REGEXES = {
    int     : r'(-?\d+)',
    float   : r'(-?\d+(?:\.\d+)?)',
}

# Formats for _parse_date to use.
DATE_FORMATS = (
    "%a, %d %b %Y %H:%M:%S %Z",
    "%A, %d-%b-%y %H:%M:%S %Z",
    "%a %b %d %H:%M:%S %Y",
    )

###############################################################################
# Special Exceptions
###############################################################################

class HTTPException(Exception):
    """
    Raising an instance of HTTPException will cause the Application to render
    an error page out to the client with the given
    `HTTP status code <http://en.wikipedia.org/wiki/List_of_HTTP_status_codes>`_,
    message, and any provided headers.

    This is, generally, preferable to allowing an exception of a different
    type to bubble up to the Application, which would result in a
    ``500 Internal Server Error`` page.

    The :func:`abort` helper function makes it easy to raise instances of
    this exception.

    =========  ============
    Argument   Description
    =========  ============
    status     *Optional.* The `HTTP status code <http://en.wikipedia.org/wiki/List_of_HTTP_status_codes>`_ to generate an error page for. If this isn't specified, a ``404 Not Found`` page will be generated.
    message    *Optional.* A text message to display on the error page.
    headers    *Optional.* A dict of extra HTTP headers to return with the rendered page.
    =========  ============
    """
    def __init__(self, status=404, message=None, headers=None):
        self.status = status
        self.message = message
        self.headers = headers

class HTTPTransparentRedirect(Exception):
    """
    Raising an instance of HTTPTransparentRedirect will cause the Application
    to silently redirect a request to a new URI.
    """
    def __init__(self, uri):
        self.uri = uri

###############################################################################
# Application Class
###############################################################################

class Application(object):
    """
    The Application class builds upon :class:`pants.contrib.http.HTTPServer`,
    adding support for request routing, additional error handling, and a
    degree of convenience that makes writing dynamic pages easier.

    Instances of Application are callable, and may be used as a HTTPServer's
    request handler.

    ===============  ============
    Argument         Description
    ===============  ============
    default_domain   *Optional.* The default domain to search for a route for if the request's Host does not exist.
    debug            *Optional.* If this is set to True, automatically generated ``500 Internal Server Error`` response pages will include information about the failed request, including a traceback of the exception that caused the page to be generated.
    ===============  ============
    """
    current_app = None

    def __init__(self, default_domain=None, debug=False):
        # Internal Stuff
        self._routes    = {}
        self._names     = {}

        self._routes[None] = {}

        # External Stuff
        self.default_domain = None
        self.debug = debug

    def run(self, port=None, host='', ssl_options=None):
        """
        This function exists for convenience, and when called creates a
        :class:`~pants.contrib.http.HTTPServer` instance with its request
        handler set to this application instance, calls
        :func:`~pants.contrib.http.HTTPServer.listen` on that HTTPServer, and
        finally, starts the Pants engine to process requests.

        ============  ============
        Argument      Description
        ============  ============
        port          *Optional.* The port to listen on. If this isn't specified, it will be either 80 or 443, depending on whether or not SSL options for the server have been provided.
        host          *Optional.* The host interface to listen on. If this isn't specified, listen on all interfaces.
        ssl_options   *Optional.* A dict of SSL options for the server. See :class:`pants.contrib.ssl.SSLServer` for more information.
        ============  ============
        """
        from pants import engine
        HTTPServer(self, ssl_options=ssl_options).listen(port, host)
        engine.start()

    ##### Route Management Methods ############################################

    def basic_route(self, rule, name=None, methods=['GET','HEAD']):
        """
        The basic_route decorator registers a route with the Application without
        holding your hand over it.

        It functions almost the same as the :func:`Application.route` decorator,
        but doesn't wrap the provided function with any argument handling code.
        Instead, you're provided with the request object and the the regex
        match object.

        Example Usage::

            @app.basic_route("/char/<char>")
            def my_route(request):
                char, = request.match.groups()
                return 'The character is %s!' % char

        That is, essentially, equivilent to::

            @app.route("/char/<char>/")
            def my_route(char):
                return 'The character is %s!' % char

        =========  ============
        Argument   Description
        =========  ============
        rule       The route rule to match for a request to go to the decorated function. See :func:`Application.route` for more information.
        name       *Optional.* The name of the decorated function, for use with the :func:`url_for` helper function.
        methods    *Optional.* A list of HTTP methods to allow for this request handler. By default, only ``GET`` and ``HEAD`` requests are allowed, and all others will result in a ``405 Method Not Allowed`` error.
        =========  ============
        """
        def decorator(func):
            if rule[0] != '/':
                domain, _, rule = rule.partition('/')
                rule = '/' + rule
            else:
                domain = None

            regex, arguments, names, namegen = _route_to_regex(rule)
            _regex = re.compile(regex)

            if name is None:
                name = "%s.%s" % (func.__module__, func.__name__)

            self._insert_route(_regex, func, domain, name, methods, names, namegen)
            return func
        return decorator

    def route(self, rule, name=None, methods=['GET','HEAD'], auto404=False):
        """
        The route decorator is used to register a new request handler with the
        Application instance. Example::

            @app.route("/")
            def hello_world():
                return "Hiya, Everyone!"

        Variables may be specified in the route *rule* by wrapping them with
        inequality signs (for example: ``<variable_name>``). By default, a
        variable part accepts any character except a slash (``/``) and returns
        a string value. However, you may specify a specific type to be returned
        by using the format ``<type:name>``, where type is the name of a
        callable in the pants.contrib.web namespace that accepts a single
        string as its argument, and returns a value. Built-in types, such as
        int and float, work well for this. Example::

            @app.route("/user/<int:id>/")
            def user(id):
                return "Hi, user %d!" % id

        The ``id`` is automatically converted into an integer for you, and as
        an added bonus, your function is never even called if the provided
        value for ``id`` isn't a valid number.

        Request handlers are easy to write and can send their output to the
        client simply by returning a value, such as a string::

            @app.route("/")
            def hello_world():
                return "Hiya, Everyone!"

        The previous code would result in a `200 OK`` response, with a
        ``Content-Type`` header of ``text/plain``, and a ``Content-Length``
        header of ``15``. With, of course, the body ``Hiya, Everyone!``.

        If the returned string begins with ``<!DOCTYPE`` or ``<html``, it will
        be assumed that the response is of ``Content-Type: text/html``.

        If a unicode object is returned, rather than a simple string, it will
        be automatically encoded and an encoding argument will be added to the
        ``Content-Type`` header.

        If a dictionary is returned, it will be automatically converted to a
        string of `JSON <http://en.wikipedia.org/wiki/JSON>`_ and the
        ``Content-Type`` header will be set to ``application/json``.

        If any other object is returned, it will be converted to a string
        via ``str()`` before any content headers are set. The exception to this
        is that, if the object has a ``__html__`` method, that method will be
        called rather than ``str()``, and the ``Content-Type`` will be
        automatically assumed to be ``text/html``, regardless of the actual
        content of the string.

        A tuple of ``(body, status)`` or ``(body, status, headers)`` may be
        returned, rather than simply a body, to set the HTTP status code of
        the result and additional response headers. If provided, ``status``
        must be an integer, and ``headers`` must be a dict.

        The following example returns a page with the status code ``404 Not Found``::

            @app.route("/nowhere")
            def nowhere():
                return "This does not exist.", 404

        =========  ============
        Argument   Description
        =========  ============
        rule       The route rule to be matched for the decorated function to be used for handling a request.
        name       *Optional.* The name of the decorated function, for use with the :func:`url_for` helper function.
        methods    *Optional.* A list of HTTP methods to allow for this request handler. By default, only ``GET`` and ``HEAD`` requests are allowed, and all others will result in a ``405 Method Not Allowed`` error.
        auto404    *Optional.* If this is set to True, all response handler arguments will be checked for truthiness (True, non-empty strings, etc.) and, if any fail, a ``404 Not Found`` page will be rendered automatically.
        =========  ============
        """
        if callable(name):
            self._add_route(rule, name, None, methods, auto404)
            return

        def decorator(func):
            self._add_route(rule, func, name, methods, auto404)
            return func
        return decorator

    ##### Error Handlers ######################################################

    def handle_404(self, request, exception):
        if isinstance(exception, HTTPException):
            return error(exception.message, 404)
        return error(404)

    def handle_500(self, request, exception):
        log.exception('Error handling HTTP request: %s %%s' % request.method,
            request.uri)
        if not self.debug:
            return error(500)

        resp = u''.join([
            u"<h2>Traceback</h2>\n",
            u"<pre>%s</pre>\n" % traceback.format_exc(),
            u"<h2>Route</h2>\n<pre>",
            u"route name   = %r\n" % request.route_name,
            u"match groups = %r" % (request.match.groups(),),
            u"</pre>\n",
            u"<h2>HTTP Request</h2>\n",
            request.__html__(),
            ])

        return error(resp, 500)

    ##### The Request Handler #################################################

    def __call__(self, request):
        """
        This function is called when a new request is received, and calls both
        :func:`Application.handle_request` and :func:`Application.handle_output`
        to process the request.
        """
        Application.current_app = self
        self.request = request

        try:
            request.auto_finish = True
            self.handle_output(self.handle_request(request))
        finally:
            request.route = None
            request.match = None
            request.route_name = None

            Application.current_app = None
            self.request = None

    def handle_output(self, result):
        """ Process the output of handle_request. """
        request = self.request

        if not request.auto_finish or result is None or \
                request._finish is not None:
            if request.auto_finish and request._finish is None:
                request.finish()
            return

        status = 200
        if type(result) is tuple:
            if len(result) == 3:
                body, status, headers = result
            else:
                body, status = result
                headers = {}
        else:
            body = result
            headers = {}

        # Set a Content-Type header if there isn't already one.
        if not 'Content-Type' in headers:
            if (isinstance(body, basestring) and
                    body[:5].lower() in ('<html','<!doc')) or \
                    hasattr(body, '__html__'):
                headers['Content-Type'] = 'text/html'
            elif isinstance(body, dict):
                headers['Content-Type'] = 'application/json'
            else:
                headers['Content-Type'] = 'text/plain'

        # Convert the body to something sendable.
        try:
            body = body.__html__()
        except AttributeError:
            pass

        if isinstance(body, unicode):
            encoding = headers['Content-Type']
            if 'charset=' in encoding:
                before, sep, enc = encoding.partition('charset=')
            else:
                before = encoding
                sep = '; charset='
                enc = 'UTF-8'

            body = body.encode(enc)
            headers['Content-Type'] = '%s%s%s' % (before, sep, enc)

        elif isinstance(body, dict):
            try:
                body = json.dumps(body)
            except Exception, e:
                body, status, headers = self.handle_500(request, e)
                body = body.encode('utf-8')
                headers['Content-Type'] = 'text/html; charset=UTF-8'

        elif not isinstance(body, str):
            body = str(body)

        # More headers!
        headers['Content-Length'] = len(body)
        if not 'Date' in headers:
            headers['Date'] = date(datetime.utcnow())
        if not 'Server' in headers:
            headers['Server'] = SERVER

        # Send the response.
        request.send_status(status)
        request.send_headers(headers)

        if request.method == 'HEAD':
            request.finish()
            return

        request.write(body)
        request.finish()

    def handle_request(self, request):
        path = request.path

        # Domain Matching
        if len(self._routes) == 1:
            domain = None
        else:
            if request.host in self._routes:
                domain = request.host
            else:
                domain = '.' + request.host.partition('.')[2]
                if not domain in self._routes and ':' in request.host:
                    domain = request.host.rpartition(':')[0]
                    if not domain in self._routes:
                        domain = '.' + domain.partition('.')[2]
                if not domain in self._routes:
                    domain = self.default_domain

        for route in self._routes[domain]:
            match = route.match(path)
            if match is None:
                continue

            # Process this route.
            func, name, methods = self._routes[domain][route][:3]

            request.route = route
            request.match = match
            request.route_name = name

            if request.method not in methods:
                return error(
                    'The method %s is not allowed for %r.' % (
                        request.method, path), 405, {
                            'Allow': ', '.join(methods)
                        })
            else:
                try:
                    return func(request)
                except HTTPException, e:
                    if hasattr(self, 'handle_%d' % e.status):
                        return getattr(self, 'handle_%d' % e.status)(request, e)
                    else:
                        return error(e.message, e.status, e.headers)
                except HTTPTransparentRedirect, e:
                    request.uri = e.uri
                    request._parse_uri()
                    return self.handle_request(request)
                except Exception, e:
                    return self.handle_500(request, e)
            break
        else:
            # No matching routes.
            if not path.endswith('/'):
                p = '%s/' % path
                for route in self._routes[domain]:
                    if route.match(p):
                        if request.query:
                            return redirect('%s?%s' % (p,request.query))
                        else:
                            return redirect(p)

        return self.handle_404(request, None)

    ##### Internal Methods and Event Handlers #################################

    def _insert_route(self, route, handler, domain, name, methods, nms, namegen):
        if isinstance(route, basestring):
            route = re.compile(route)
        if not domain in self._routes:
            self._routes[domain] = {}
        self._routes[domain][route] = (handler, name, methods, nms, namegen)
        self._names[name] = route

    def _add_route(self, route, view, name=None, methods=['GET','HEAD'],
            auto404=False):
        """ See: Application.route """
        if name is None:
            if view is None:
                raise Exception('No name or view specified!')
            if hasattr(view, '__name__'):
                name = view.__name__
            elif hasattr(view, '__class__'):
                name = view.__class__.__name__
            else:
                raise NameError("Cannot find name for this route.")

        if not callable(view):
            raise Exception('View must be callable.')

        # Parse the route.
        if route[0] != '/':
            domain, _, route = route.partition('/')
            route = '/' + route
        else:
            domain = None

        regex, arguments, names, namegen = _route_to_regex(route)
        _regex = re.compile(regex)

        if not arguments:
            arguments = False

        try:
            args = inspect.getargspec(view).args
        except TypeError:
            args = inspect.getargspec(view.__call__).args[1:]

        if len(args) == 1 and args[0] == 'request':
            def view_runner(request):
                request.__viewmodule__ = view.__module__
                match = request.match
                try:
                    if arguments is False:
                        return view(request)

                    out = []
                    for val,type in zip(match.groups(), arguments):
                        if type is not None:
                            try:
                                val = type(val)
                            except Exception:
                                return error('Unable to parse data %r.' % val, 400)
                        out.append(val)

                    if auto404 is True:
                        all_or_404(*out)

                    request.arguments = out
                    return view(request)
                finally:
                    request.arguments = None

        else:
            def view_runner(request):
                request.__viewmodule__ = view.__module__
                match = request.match
                try:
                    try:
                        view.func_globals['request'] = request
                    except AttributeError:
                        view.__call__.func_globals['request'] = request
                    if arguments is False:
                        return view()

                    out = []
                    for val,type in zip(match.groups(), arguments):
                        if type is not None:
                            try:
                                val = type(val)
                            except Exception:
                                return error('Unable to parse data %r.' % val, 400)
                        out.append(val)

                    if auto404 is True:
                        all_or_404(*out)

                    return view(*out)
                finally:
                    try:
                        view.func_globals['request'] = None
                    except AttributeError:
                        view.__call__.func_globals['request'] = None

        view_runner.__name__ = name
        self._insert_route(_regex, view_runner, domain,
            "%s.%s" %(view.__module__,name), methods, names, namegen)

###############################################################################
# FileServer Class
###############################################################################

class FileServer(object):
    """
    The FileServer is a request handling class that, as it sounds, serves files
    to the client. It also supports the ``Content-Range`` header, HEAD requests,
    and last modified dates.

    ==========  ==============================  ============
    Argument    Default                         Description
    ==========  ==============================  ============
    path                                        The path to serve.
    blacklist   ``.py`` and ``.pyc`` files      *Optional.* A list of regular expressions to test filenames against. If a given file matches any of the provided patterns, it will not be downloadable and instead return a ``403 Unauthorized`` error.
    default     ``index.html``, ``index.htm``   *Optional.* A list of default files to be displayed rather than a directory listing if they exist.
    renderers   None                            *Optional.* A dictionary of methods for rendering files with a given extension into more suitable output, such as converting rST to HTML, or minifying CSS.
    ==========  ==============================  ============

    It attempts to serve the files as efficiently as possible, using the
    `sendfile <http://www.kernel.org/doc/man-pages/online/pages/man2/sendfile.2.html>`_
    system call when possible, and with proper use of ETags and other headers to
    minimize repetitive downloading.

    Using it is simple. It only requires a single argument: the path to serve
    files from. You can also supply a list of default files to check to serve
    rather than a file listing.

    When used with an Application, the FileServer is not created in the usual
    way with the route decorator, but rather with a method of the FileServer
    itself. Example::

        FileServer("/tmp/path").attach(app)

    If you wish to listen on a path other than ``/static/``, you can also use
    that when attaching::

        FileServer("/tmp/path").attach(app, "/files/")
    """
    def __init__(self, path, blacklist=[re.compile('.*\.pyc?$')],
            defaults=['index.html','index.htm'],
            renderers=None):
        # Make sure our path is unicode.
        if not isinstance(path, unicode):
            path = _decode(path)

        self.path = os.path.normpath(os.path.realpath(path))
        self.defaults = defaults
        self.renderers = renderers or {}

        # Build the blacklist.
        self.blacklist = []
        for bl in blacklist:
            if isinstance(bl, str):
                bl = re.compile(bl)
            self.blacklist.append(bl)

    def attach(self, app, path='/static/', domain=None):
        """
        Attach this fileserver to an application, bypassing the usual route
        decorator to ensure things are done right.

        =========  ===============  ============
        Argument   Default          Description
        =========  ===============  ============
        app                         The :class:`~pants.contrib.web.Application` instance to attach to.
        path       ``'/static/'``   *Optional.* The path to serve requests from.
        domain     None             *Optional.* The domain to serve requests upon.
        =========  ===============  ============
        """
        path = re.compile("^%s(.*)$" % re.escape(path))
        app._insert_route(path, self, domain, "FileServer", ['HEAD','GET'], None, None)

    def check_blacklist(self, path):
        """
        Check the given path to make sure it isn't blacklisted. If it is
        blacklisted, then raise an :class:`~pants.contrib.web.HTTPException`
        via :func:`~pants.contrib.web.abort`.

        =========  ============
        Argument   Description
        =========  ============
        path       The path to check against the blacklist.
        =========  ============
        """
        for bl in self.blacklist:
            if isinstance(bl, unicode):
                if bl in path:
                    abort(403)
            elif bl.match(path):
                abort(403)

    def __call__(self, request):
        """
        Serve a request.
        """

        try:
            path = request.match.group(1)
        except (AttributeError, IndexError):
            path = request.path

        # Conver the path to unicode.
        path = _decode(urllib.unquote(path))

        # Strip off a starting quote.
        if path.startswith('/') or path.startswith('\\'):
            path = path[1:]

        # Normalize the path.
        full_path = os.path.normpath(os.path.join(self.path, path))

        # Validate the request.
        if not full_path.startswith(self.path):
            abort(403)
        if not os.path.exists(full_path):
            abort(404)

        # Is this a directory?
        if os.path.isdir(full_path):
            # Check defaults.
            for f in self.defaults:
                full = os.path.join(full_path, f)
                if os.path.exists(full):
                    request.path = urllib.quote(full.encode('utf8'))
                    if hasattr(request, 'match'):
                        del request.match
                    return self.__call__(request)

            # Guess not. List it.
            if hasattr(request, 'match'):
                return self.list_directory(request, path)
            else:
                body, status, headers = self.list_directory(request, path)
                if isinstance(body, unicode):
                    body = body.encode('utf-8')
                headers['Content-Length'] = len(body)
                request.send_status(status)
                request.send_headers(headers)
                request.send(body)
                request.finish()
                return

        # Blacklist Checking.
        self.check_blacklist(full_path)

        # Try rendering the content.
        ext = os.path.basename(full_path).rpartition('.')[-1]
        if ext in self.renderers:
            f, mtime, size, type = self.renderers[ext](request, full_path)
        else:
            # Get the information for the actual file.
            f = None
            stat = os.stat(full_path)
            mtime = stat.st_mtime
            size = stat.st_size
            type = mimetypes.guess_type(full_path)[0]

        # If we don't have a type, text/plain it.
        if type is None:
            type = 'text/plain'

        # Generate a bunch of data for headers.
        modified = datetime.fromtimestamp(mtime)
        expires = datetime.utcnow() + timedelta(days=7)

        etag = '"%x-%x"' % (size, int(mtime))

        headers = {
            'Last-Modified' : date(modified),
            'Expires'       : date(expires),
            'Cache-Control' : 'max-age=604800',
            'Content-Type'  : type,
            'Date'          : date(datetime.utcnow()),
            'Server'        : SERVER,
            'Accept-Ranges' : 'bytes',
            'ETag'          : etag
        }

        do304 = False

        if 'If-Modified-Since' in request.headers:
            try:
                since = _parse_date(request.headers['If-Modified-Since'])
            except ValueError:
                since = None
            if since and since >= modified:
                do304 = True

        if 'If-None-Match' in request.headers:
            if etag == request.headers['If-None-Match']:
                do304 = True

        if do304:
            if f:
                f.close()
            request.auto_finish = False
            request.send_status(304)
            request.send_headers(headers)
            request.finish()
            return

        if 'If-Range' in request.headers:
            if etag != request.headers['If-Range'] and \
                    'Range' in request.headers:
                del request.headers['Range']

        last = size - 1
        range = 0, last
        status = 200

        if 'Range' in request.headers:
            if request.headers['Range'].startswith('bytes='):
                try:
                    val = request.headers['Range'][6:].split(',')[0]
                    start, end = val.split('-')
                except ValueError:
                    if f:
                        f.close()
                    abort(416)
                try:
                    if end and not start:
                        end = last
                        start = last - int(end)
                    else:
                        start = int(start or 0)
                        end = int(end or last)

                    if start < 0 or start > end or end > last:
                        if f:
                            f.close()
                        abort(416)
                    range = start, end
                except ValueError:
                    pass
                if range[0] != 0 or range[1] != last:
                    status = 206
                    headers['Content-Range'] = 'bytes %d-%d/%d' % (
                        range[0], range[1], size)

        # Set the content length header.
        if range[0] == range[1]:
            headers['Content-Length'] = 0
        else:
            headers['Content-Length'] = 1 + (range[1] - range[0])

        # Send the headers and status line.
        request.auto_finish = False
        request.send_status(status)
        request.send_headers(headers)

        # Don't send the body if this is head.
        if request.method == 'HEAD':
            if f:
                f.close()
            request.finish()
            return

        # Open the file and send it.
        if range[0] == range[1]:
            if f:
                f.close()
            request.finish()
            return

        if f is None:
            f = open(full_path, 'rb')

        if range[1] != last:
            length = 1 + (range[1] - range[0])
        else:
            length = 0

        def on_write():
            del request.connection.handle_write_file
            request.finish()

        request.connection.handle_write_file = on_write
        request.connection.write_file(f, nbytes=length, offset=range[0])

    def list_directory(self, request, path):
        """
        Generate a directory listing and return it.
        """

        # Normalize the path.
        full_path = os.path.normpath(os.path.join(self.path, path))

        # Get the URI, which is just request.path decoded.
        uri = _decode(urllib.unquote(request.path))
        if not uri.startswith(u'/'):
            uri = u'/%s' % uri
        if not uri.endswith(u'/'):
            return redirect(u'%s/' % uri)

        go_up = u''
        url = uri.strip(u'/')
        if url:
            go_up = u'<p><a href="..">Up to Higher Directory</a></p>'

        files = []
        dirs = []

        for p in sorted(os.listdir(full_path), key=unicode.lower):
            if _is_hidden(p, full_path):
                continue

            full = os.path.join(full_path, p)
            try:
                fp = full
                if os.path.isdir(full):
                    fp += '/'
                self.check_blacklist(fp)
            except HTTPException:
                continue

            stat = os.stat(full)
            mtime = datetime.fromtimestamp(stat.st_mtime).strftime(
                u'<td class="right">%Y-%m-%d</td>'
                u'<td class="left">%I:%M:%S %p</td>'
                )

            if os.path.isdir(full):
                cls = u'folder'
                link = u'%s/' % p
                size = u'<span class="faint">Directory</span>'
                obj = dirs

            elif os.path.isfile(full):
                cls = 'document'
                ext = p[p.rfind('.')+1:]
                if ext in ('jpg','jpeg','png','gif','bmp'):
                    cls = 'image'
                elif ext in ('zip','gz','tar','7z','tgz'):
                    cls = 'zip'
                elif ext in ('mp3','mpa','wma','wav','flac','mid','midi','raw',
                        'mod','xm','aac','m4a','ogg','aiff','au','voc','m3u',
                        'pls','asx'):
                    cls = 'audio'
                elif ext in ('mpg','mpeg','mkv','mp4','wmv','avi','mov'):
                    cls = 'video'
                link = p
                size = _human_readable_size(stat.st_size)
                obj = files

            else:
                continue

            obj.append(
                u'<tr><td><a class="icon %s" href="%s%s">%s</a></td><td>%s'
                u'</td>%s</tr>' % (
                    cls, uri, link, p, size, mtime))

        if files or dirs:
            files = u''.join(dirs) + u''.join(files)
        else:
            files = (u'<tr><td colspan="4" class="noborder">'
                     u'<div class="footer center">'
                     u'This directory is empty.</div></td></tr>')

        if Application.current_app and Application.current_app.debug:
            rtime = u'%0.3f ms' % (1000 * request.time)
        else:
            rtime = u''

        return DIRECTORY_PAGE % (uri, uri, go_up, files, request.host, rtime), \
            200, {
                'Content-Type':'text/html; charset=utf-8'
            }

###############################################################################
# Private Helper Functions
###############################################################################

def path(st):
    return st
path.regex = "(.+?)"

def _get_thing(thing):
    if thing in globals():
        return globals()[thing]
    elif type(__builtins__) is dict and thing in __builtins__:
        return __builtins__[thing]
    elif hasattr(__builtins__, thing):
        return getattr(__builtins__, thing)
    return None

_route_parser = re.compile(r"<([^>]+)>([^<]*)")
def _route_to_regex(route):
    """ Parse a Flask-style route and return a regular expression, as well as
        a tuple of things for conversion. """
    regex, values, names, namegen = "", [], [], ""
    if not route.startswith("^/"):
        if route.startswith("/"):
            route = "^%s$" % route
        else:
            route = "^/%s$" % route

    # Find up to the first < and add it to regex.
    ind = route.find('<')
    if ind is -1:
        return route, tuple(), tuple(), route[1:-1]
    elif ind > 0:
        regex += route[:ind]
        namegen += route[:ind]
        route = route[ind:]

    # If the parser doesn't match, return.
    if not _route_parser.match(route):
        return regex+route, tuple(), tuple(), (regex+route)[1:-1]

    for match in _route_parser.finditer(route):
        group = match.group(1)
        if ':' in group:
            type, var = group.split(':', 1)
            thing = _get_thing(type)
            if not thing:
                raise Exception, "Invalid type declaration, %s" % type
            if hasattr(thing, 'regex'):
                regex += thing.regex
            elif thing in REGEXES:
                regex += REGEXES[thing]
            else:
                regex += "([^/]+)"
            values.append(thing)
            names.append(var)
        else:
            regex += "([^/]+)"
            values.append(None)
            names.append(group)
        namegen += "%s" + match.group(2)
        regex += match.group(2)

    return regex, tuple(values), tuple(names), namegen[1:-1]

_abbreviations = (
    (1<<50L, u' PB'),
    (1<<40L, u' TB'),
    (1<<30L, u' GB'),
    (1<<20L, u' MB'),
    (1<<10L, u' KB'),
    (1, u' B')
)
def _human_readable_size(size, precision=2):
    """ Convert a size to a human readable filesize. """
    if size == 0:
        return u'0 B'

    for f,s in _abbreviations:
        if size >= f:
            break

    ip, dp = `size/float(f)`.split('.')
    if int(dp[:precision]):
        return  u'%s.%s%s' % (ip,dp[:precision],s)
    return u'%s%s' % (ip,s)

_encodings = ('utf-8','iso-8859-1','cp1252','latin1')
def _decode(text):
    for enc in _encodings:
        try:
            return text.decode(enc)
        except UnicodeDecodeError:
            continue
    else:
        return text.decode('utf-8','ignore')

def _parse_date(text):
    for fmt in DATE_FORMATS:
        try:
            return datetime(*time.strptime(text, fmt)[:6])
        except ValueError:
            continue
    raise ValueError("Unable to parse time data %r." % text)

###############################################################################
# Public Helper Functions
###############################################################################

def abort(status=404, message=None, headers=None):
    """
    Raise a :class:`~pants.contrib.web.HTTPException` to display an error page.
    """
    raise HTTPException(status, message, headers)

def all_or_404(*args):
    """
    If any of the provided arguments aren't truthy, raise a ``404 Not Found``
    exception. This is automatically called for you if you set ``auto404=True``
    when using the route decorator.
    """
    all(args) or abort()

def error(message=None, status=None, headers=None, request=None, debug=None):
    """
    Return a very simple error page, defaulting to a ``404 Not Found`` error if
    no status code is supplied. Usually, you'll want to call :func:`~pants.contrib.web.abort`
    in your code, rather than error(), to streamline the process of abandoning
    your code. Usage::

        return error(404)
        return error("Some message.", 404)
        return error("Blah blah blah.", 403, {'Some-Header':'Fish'})
    """
    if request is None:
        request = Application.current_app.request

    if status is None:
        if type(message) is int:
            status = message
            message = None
        else:
            status = 404

    if not status in HTTP:
        status = 404
    title = HTTP[status]
    if not headers:
        headers = {}

    if message is None:
        if status in HTTP_MESSAGES:
            dict = request.__dict__.copy()
            dict['uri'] = _decode(urllib.unquote(dict['uri']))
            message = HTTP_MESSAGES[status] % dict
        else:
            message = u"An unspecified error has occured."

    haiku = u''
    if status in HAIKUS:
        haiku = u'<div class="haiku">%s</div>' % HAIKUS[status]

    if not message.startswith(u'<'):
        message = u'<p>%s</p>' % message

    if debug is None:
        debug = Application.current_app and Application.current_app.debug

    if debug:
        time = u'%0.3f ms' % (1000 * request.time)
    else:
        time = u''

    result = ERROR_PAGE % (status, title, status, title.replace(u' ',u'&nbsp;'),
        haiku, message, request.host, time)

    return result, status, headers

def redirect(uri, status=302):
    """
    Construct a ``302 Found`` response to instruct the client's browser to
    redirect its request to a different URL. Other codes may be returned by
    specifying a status.

    =========  ========  ============
    Argument   Default   Description
    =========  ========  ============
    uri                  The URI to redirect the client's browser to.
    status     ``302``   *Optional.* The status code to send with the response.
    =========  ========  ============
    """
    url = uri
    if isinstance(url, unicode):
        url = uri.encode('utf-8')

    return error(
        'The document you have requested is located at <a href="%s">%s</a>.' % (
            uri, uri), status, {'Location':url})

def url_for(name, **values):
    """
    Generates a URL to the route with the given name. The name is relative to
    the module of the route function. Examples:

    ==============  ================  ================
    View's Module   Target Endpoint   Target Function
    ==============  ================  ================
    ``test``        ``index``         ``test.index``
    ``test``        ``.who``          The first ``who`` function in *any* module.
    ``test``        ``admin.login``   ``admin.login``
    ==============  ================  ================

    Any value provided to the function with an unknown key is appended to the
    generated URL as query arguments. For example, take the following route::

        @app.route("/user/<int:id>/")
        def user_page(id):
            pass

    Assuming ``url_for`` is used within the same module, the following examples
    will hold true::

        >>> url_for("user_page", id=12)
        '/user/12/'

        >>> url_for("user_page", id=12, section=3)
        '/user/12/?section=3'

        >>> url_for("user_page", id=12, _external=True)
        'http://www.example.com/user/12/'

    As demonstrated above, the ``_external`` parameter is special, and will
    result in the generation of a full URL, using the scheme and host provided
    by the current request.

    *Note:* This function has not yet been updated to properly make use of
    domains.
    """
    app = Application.current_app
    request = app.request

    if name.startswith('.'):
        # Find it in the first possible place.
        name = name[1:]
        for n in app._names:
            module, nm = n.split('.',1)
            if nm == name:
                name = n
                break
    elif not '.' in name:
        # Find it in this module.
        name = "%s.%s" % (request.__viewmodule__, name)

    if not name in app._names:
        raise NameError("Cannot find route %r." % name)

    route = app._names[name]
    names, namegen = app._routes[route][-2:]

    out = []
    for n in names:
        out.append(str(values[n]))
        del values[n]
    out = tuple(out)

    if len(out) == 1:
        out = namegen % out[0]
    else:
        out = namegen % out
    out = urllib.quote(out)

    if '_external' in values:
        if values['_external']:
            out = '%s://%s%s' % (request.protocol, request.host, out)
        del values['_external']

    if values:
        out += '?%s' % urllib.urlencode(values)

    return out



###############################################################################
#
# Copyright 2011 Pants Developers (see AUTHORS.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################
"""
A convenient collection of high-level network channels.
"""

###############################################################################
# Imports
###############################################################################

import weakref

from pants.stream import Stream, StreamServer


###############################################################################
# Client Class
###############################################################################

class Client(Stream):
    """
    A network socket client.
    """
    def __init__(self):
        # This dummy method prevents keyword arguments from finding
        # their way up to the Stream/_Channel constructors.
        Stream.__init__(self)

    ##### Control Methods #####################################################

    def connect(self, host, port):
        """
        Connect the channel to a remote socket.

        Returns the channel.

        ==========  ============
        Arguments   Description
        ==========  ============
        host        The remote host to connect to.
        port        The port to connect on.
        ==========  ============
        """
        return Stream.connect(self, (host, port))


###############################################################################
# Connection Class
###############################################################################

class Connection(Stream):
    """
    A connection to a network socket server.

    =========  ============
    Argument   Description
    =========  ============
    socket     A pre-existing socket that this channel should wrap.
    server     The server to which this channel is connected.
    =========  ============
    """
    def __init__(self, socket, server):
        Stream.__init__(self, socket=socket)

        self.server = server


###############################################################################
# Server Class
###############################################################################

class Server(StreamServer):
    """
    A network socket server.

    ================  ============
    Argument          Description
    ================  ============
    ConnectionClass   *Optional.* A :obj:`pants.network.Connection` subclass with which to wrap newly connected sockets.
    ================  ============
    """
    #: A :obj:`pants.network.Connection` subclass with which to wrap newly connected sockets.
    ConnectionClass = Connection

    def __init__(self, ConnectionClass=None):
        StreamServer.__init__(self)

        # Sets instance attribute, NOT class attribute.
        if ConnectionClass:
            self.ConnectionClass = ConnectionClass

        self.channels = weakref.WeakValueDictionary()  # fd : channel

    ##### Control Methods #####################################################

    def listen(self, port=8080, host='', backlog=1024):
        """
        Begin listening for connections made to the channel.

        Returns the channel.

        ==========  ============
        Arguments   Description
        ==========  ============
        port        *Optional.* The port to listen for connection on. By default, is 8080.
        host        *Optional.* The local host to bind to. By default, is ''.
        backlog     *Optional.* The size of the connection queue. By default, is 1024.
        ==========  ============
        """
        return StreamServer.listen(self, (host, port), backlog)

    ##### Public Event Handlers ###############################################

    def on_accept(self, socket, addr):
        """
        Called after the channel has accepted a new connection.

        Create a new instance of :attr:`ConnectonClass` to wrap the socket
        and add it to the server.

        =========  ============
        Argument   Description
        =========  ============
        sock       The newly connected socket object.
        addr       The new socket's address.
        =========  ============
        """
        connection = self.ConnectionClass(socket, self)
        self.channels[connection.fileno] = connection
        connection._handle_connect_event()

    def on_close(self):
        """
        Called after the channel has finished closing.

        Close all active connections to the server.
        """
        for channel in self.channels.values():
            channel.close()
